#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from __future__ import annotations

import abc
import logging
from itertools import combinations
from typing import TYPE_CHECKING, Any, List, Optional

try:
    from rapidfuzz import fuzz, utils

    IS_RAPIDFUZZ_INSTALLED = True
except ImportError:
    IS_RAPIDFUZZ_INSTALLED = False

try:
    import numpy as np
    import spacy
    from spacy.cli.download import download as spacy_download
    from spacy.language import Language

    IS_SPACY_INSTALLED = True
except Exception:
    IS_SPACY_INSTALLED = False


if TYPE_CHECKING:
    import numpy as np
    from numpy.typing import NDArray

import neo4j

from neo4j_graphrag.experimental.components.types import ResolutionStats
from neo4j_graphrag.experimental.pipeline import Component
from neo4j_graphrag.experimental.pipeline.component import ComponentMeta
from neo4j_graphrag.utils import driver_config

logger = logging.getLogger(__name__)


class EntityResolver(Component):
    """Entity resolution base class

    Args:
        driver (neo4j.Driver): The Neo4j driver to connect to the database.
        filter_query (Optional[str]): Cypher query to select the entities to resolve. By default, all nodes with __Entity__ label are used
    """

    def __init__(
        self,
        driver: neo4j.Driver,
        filter_query: Optional[str] = None,
    ) -> None:
        self.driver = driver_config.override_user_agent(driver)
        self.filter_query = filter_query

    async def run(self, *args: Any, **kwargs: Any) -> ResolutionStats:
        raise NotImplementedError()


class SinglePropertyExactMatchResolver(EntityResolver):
    """Resolve entities with same label and exact same property (default is "name").

    Args:
        driver (neo4j.Driver): The Neo4j driver to connect to the database.
        filter_query (Optional[str]): To reduce the resolution scope, add a Cypher WHERE clause.
        resolve_property (str): The property that will be compared (default: "name"). If values match exactly, entities are merged.
        neo4j_database (Optional[str]): The name of the Neo4j database. If not provided, this defaults to the server's default database ("neo4j" by default) (`see reference to documentation <https://neo4j.com/docs/operations-manual/current/database-administration/#manage-databases-default>`_).

    Example:

    .. code-block:: python

        from neo4j import GraphDatabase
        from neo4j_graphrag.experimental.components.resolver import SinglePropertyExactMatchResolver

        URI = "neo4j://localhost:7687"
        AUTH = ("neo4j", "password")
        DATABASE = "neo4j"

        driver = GraphDatabase.driver(URI, auth=AUTH)
        resolver = SinglePropertyExactMatchResolver(driver=driver, neo4j_database=DATABASE)
        await resolver.run()  # no expected parameters

    """

    def __init__(
        self,
        driver: neo4j.Driver,
        filter_query: Optional[str] = None,
        resolve_property: str = "name",
        neo4j_database: Optional[str] = None,
    ) -> None:
        super().__init__(driver, filter_query)
        self.resolve_property = resolve_property
        self.neo4j_database = neo4j_database

    async def run(self) -> ResolutionStats:
        """Resolve entities based on the following rule:
        For each entity label, entities with the same 'resolve_property' value
        (exact match) are grouped into a single node:

        - Properties: the property from the first node will remain if already set, otherwise the first property in list will be written.
        - Relationships: merge relationships with same type and target node.

        See apoc.refactor.mergeNodes documentation for more details.
        """
        match_query = "MATCH (entity:__Entity__) "
        if self.filter_query:
            match_query += self.filter_query
        stat_query = f"{match_query} RETURN count(entity) as c"
        records, _, _ = self.driver.execute_query(
            stat_query,
            database_=self.neo4j_database,
        )
        number_of_nodes_to_resolve = records[0].get("c")
        if number_of_nodes_to_resolve == 0:
            return ResolutionStats(
                number_of_nodes_to_resolve=0,
            )
        merge_nodes_query = (
            f"{match_query} "
            f"WITH entity, entity.{self.resolve_property} as prop "
            # keep only entities for which the resolve_property (name) is not null
            "WITH entity, prop WHERE prop IS NOT NULL "
            # will check the property for each of the entity labels,
            # except the reserved ones __Entity__ and __KGBuilder__
            "UNWIND labels(entity) as lab  "
            "WITH lab, prop, entity WHERE NOT lab IN ['__Entity__', '__KGBuilder__'] "
            # aggregate based on property value and label
            # collect all entities with exact same property and label
            # in the 'entities' list
            "WITH prop, lab, collect(entity) AS entities "
            # merge all entities into a single node
            # * merge relationships: if the merged entities have a relationship of same
            # type to the same target node, these relationships are merged
            # otherwise relationships are just attached to the newly created node
            # * properties: if the two entities have the same property key with
            # different values, only one of them is kept in the created node
            "CALL apoc.refactor.mergeNodes(entities,{ "
            " properties:'discard', "
            " mergeRels:true "
            "}) "
            "YIELD node "
            "RETURN count(node) as c "
        )
        records, _, _ = self.driver.execute_query(
            merge_nodes_query, database_=self.neo4j_database
        )
        number_of_created_nodes = records[0].get("c")
        return ResolutionStats(
            number_of_nodes_to_resolve=number_of_nodes_to_resolve,
            number_of_created_nodes=number_of_created_nodes,
        )


class CombinedMeta(ComponentMeta, abc.ABCMeta):
    """
    A metaclass that merges ComponentMeta (from Component) and ABCMeta (from abc.ABC).
    """

    pass


class BasePropertySimilarityResolver(EntityResolver, abc.ABC, metaclass=CombinedMeta):
    """
    Base class for similarity-based matching of properties for entity resolution.
    Resolve entities with same label and similar set of textual properties (default is
    ["name"]):
    - Group entities by label
    - Concatenate the specified textual properties
    - Compute similarity between each pair
    - Consolidate overlapping sets
    - Merge similar nodes via APOC (See apoc.refactor.mergeNodes documentation for more
    details).

    Subclasses implement `compute_similarity` based on different strategies, and return
    a similarity score between 0 and 1.

    Args:
        driver (neo4j.Driver): The Neo4j driver to connect to the database.
        filter_query (Optional[str]): Optional Cypher WHERE clause to reduce the resolution scope.
        resolve_properties (Optional[List[str]]): The list of properties to consider for similarity. Defaults to ["name"].
        similarity_threshold (float): The similarity threshold above which nodes are merged. Defaults to 0.8.
        neo4j_database (Optional[str]): The name of the Neo4j database. If not provided, this defaults to the server's default database ("neo4j" by default) (`see reference to documentation <https://neo4j.com/docs/operations-manual/current/database-administration/#manage-databases-default>`_).

    """

    def __init__(
        self,
        driver: neo4j.Driver,
        filter_query: Optional[str] = None,
        resolve_properties: Optional[List[str]] = None,
        similarity_threshold: float = 0.8,
        neo4j_database: Optional[str] = None,
    ) -> None:
        super().__init__(driver, filter_query)
        self.resolve_properties = resolve_properties or ["name"]
        self.similarity_threshold = similarity_threshold
        self.neo4j_database = neo4j_database

    @abc.abstractmethod
    def compute_similarity(self, text_a: str, text_b: str) -> float:
        """
        Compute similarity between two textual strings.
        """
        pass

    async def run(self) -> ResolutionStats:
        match_query = "MATCH (entity:__Entity__)"
        if self.filter_query:
            match_query += f" {self.filter_query}"

        # generate a dynamic map of requested properties, e.g. "name: entity.name, description: entity.description, ..."
        props_map_list = [f"{prop}: entity.{prop}" for prop in self.resolve_properties]
        props_map = ", ".join(props_map_list)

        # Cypher query:
        # matches extracted entities
        # filters entities if filter_query is provided
        # unwinds labels to skip reserved ones
        # collects all properties needed for the calculation of similarity
        query = f"""
            {match_query}
            UNWIND labels(entity) AS lab
            WITH lab, entity
            WHERE NOT lab IN ['__Entity__', '__KGBuilder__']
            WITH lab, collect({{ id: elementId(entity), {props_map} }}) AS labelCluster
            RETURN lab, labelCluster
        """

        records, _, _ = self.driver.execute_query(query, database_=self.neo4j_database)

        total_entities = 0
        total_merged_nodes = 0

        # for each row, 'lab' is the label, 'labelCluster' is a list of dicts (id + textual properties)
        for row in records:
            entities = row["labelCluster"]

            node_texts = {}
            for ent in entities:
                # concatenate all textual properties (if non-null) into a single string
                texts = [
                    str(ent[p]) for p in self.resolve_properties if p in ent and ent[p]
                ]
                combined_text = " ".join(texts).strip()
                if combined_text:
                    node_texts[ent["id"]] = combined_text
            total_entities += len(node_texts)

            # compute pairwise similarity and mark those above the threshold
            pairs_to_merge = []
            for (id1, text1), (id2, text2) in combinations(node_texts.items(), 2):
                sim = self.compute_similarity(text1, text2)
                if sim >= self.similarity_threshold:
                    pairs_to_merge.append({id1, id2})

            # consolidate overlapping pairs into unique merge sets.
            merged_sets = self._consolidate_sets(pairs_to_merge)

            # perform merges in the db using APOC.
            merged_count = 0
            for node_id_set in merged_sets:
                if len(node_id_set) > 1:
                    merge_query = (
                        "MATCH (n) WHERE elementId(n) IN $ids "
                        "WITH collect(n) AS nodes "
                        "CALL apoc.refactor.mergeNodes(nodes, {properties: 'discard', mergeRels: true}) "
                        "YIELD node RETURN elementId(node)"
                    )
                    result, _, _ = self.driver.execute_query(
                        merge_query,
                        {"ids": list(node_id_set)},
                        database_=self.neo4j_database,
                    )
                    merged_count += len(result)
            total_merged_nodes += merged_count

        return ResolutionStats(
            number_of_nodes_to_resolve=total_entities,
            number_of_created_nodes=total_merged_nodes,
        )

    @staticmethod
    def _consolidate_sets(pairs: List[set[str]]) -> List[set[str]]:
        """Consolidate overlapping sets of node pairs into unique sets."""
        consolidated: List[set[str]] = []
        for pair in pairs:
            merged = False
            for cons in consolidated:
                # if there is any intersection, unify them
                if pair & cons:
                    cons.update(pair)
                    merged = True
                    break
            if not merged:
                consolidated.append(set(pair))
        return consolidated


class SpaCySemanticMatchResolver(BasePropertySimilarityResolver):
    """
    Resolve entities with same label and similar set of textual properties (default is
    ["name"]) based on spaCy's static embeddings and cosine similarities.

    Args:
        driver (neo4j.Driver): The Neo4j driver to connect to the database.
        filter_query (Optional[str]): Optional Cypher WHERE clause to reduce the resolution scope.
        resolve_properties (Optional[List[str]]): The list of properties to consider for embeddings Defaults to ["name"].
        similarity_threshold (float): The similarity threshold above which nodes are merged. Defaults to 0.8.
        spacy_model (str): The name of the spaCy model to load. Defaults to "en_core_web_lg".
        neo4j_database (Optional[str]): The name of the Neo4j database. If not provided, this defaults to the server's default database ("neo4j" by default) (`see reference to documentation <https://neo4j.com/docs/operations-manual/current/database-administration/#manage-databases-default>`_).

    Example:

    .. code-block:: python

        from neo4j import GraphDatabase
        from neo4j_graphrag.experimental.components.resolver import SpaCySemanticMatchResolver

        URI = "neo4j://localhost:7687"
        AUTH = ("neo4j", "password")
        DATABASE = "neo4j"

        driver = GraphDatabase.driver(URI, auth=AUTH)
        resolver = SpaCySemanticMatchResolver(driver=driver, neo4j_database=DATABASE)
        await resolver.run()  # no expected parameters

    """

    def __init__(
        self,
        driver: neo4j.Driver,
        filter_query: Optional[str] = None,
        resolve_properties: Optional[List[str]] = None,
        similarity_threshold: float = 0.8,
        nlp: Any = None,
        spacy_model: str = "en_core_web_lg",
        auto_download_spacy_model: bool = True,
        neo4j_database: Optional[str] = None,
    ) -> None:
        super().__init__(
            driver,
            filter_query,
            resolve_properties,
            similarity_threshold,
            neo4j_database,
        )
        self.nlp = nlp or self._load_or_download_spacy_model(
            spacy_model, auto_download_spacy_model=auto_download_spacy_model
        )
        self.embedding_cache: dict[str, NDArray[np.float64]] = {}

    async def run(self) -> ResolutionStats:
        return await super().run()

    def compute_similarity(self, text_a: str, text_b: str) -> float:
        emb1 = self._get_embedding(text_a)
        emb2 = self._get_embedding(text_b)
        sim = self._cosine_similarity(
            np.asarray(emb1, dtype=np.float64), np.asarray(emb2, dtype=np.float64)
        )
        return sim

    def _get_embedding(self, text: str) -> NDArray[np.float64]:
        if text not in self.embedding_cache:
            embedding = np.asarray(self.nlp(text).vector, dtype=np.float64)
            self.embedding_cache[text] = embedding
        return self.embedding_cache[text]

    @staticmethod
    def _cosine_similarity(
        vec1: NDArray[np.float64], vec2: NDArray[np.float64]
    ) -> float:
        """Calculate cosine similarity between two embedding vectors."""
        dot_product = np.dot(vec1, vec2)
        norm1 = np.linalg.norm(vec1)
        norm2 = np.linalg.norm(vec2)
        if not norm1 or not norm2:
            return 0.0
        return float(dot_product / (norm1 * norm2))

    @staticmethod
    def _load_or_download_spacy_model(
        model_name: str, *, auto_download_spacy_model: bool = True
    ) -> Language:
        """
        Attempt to load the specified spaCy model by name.
        If not installed, automatically download and then load it.
        """
        try:
            if not IS_SPACY_INSTALLED:
                raise ImportError("""`spacy` python module needs to be installed to use
                the SpaCySemanticMatchResolver. Install it with:
                `pip install "neo4j-graphrag[nlp]"`
                . SpaCy is not supported on Python 3.14 and later.
                """)
            return spacy.load(model_name)
        except OSError as e:
            if "doesn't seem to be a Python package or a valid path" in str(e):
                if not auto_download_spacy_model:
                    raise OSError(
                        f"spaCy model '{model_name}' is not installed. "
                        "Install it manually (e.g. `python -m spacy download "
                        f"{model_name}`) or pass a pre-loaded `nlp` instance."
                    ) from e
                logger.info(f"Model '{model_name}' not found. Downloading...")
                try:
                    spacy_download(model_name)
                except SystemExit as se:
                    raise OSError(
                        f"Failed to auto-download spaCy model '{model_name}'. "
                        "Install it manually (e.g. `python -m spacy download "
                        f"{model_name}`) or pass a pre-loaded `nlp` instance."
                    ) from se
                return spacy.load(model_name)
            else:
                raise e


class FuzzyMatchResolver(BasePropertySimilarityResolver):
    """
    Resolve entities with the same label and similar set of textual properties using
    RapidFuzz for fuzzy matching. Similarity scores are normalized to a value between 0
    and 1.
    """

    def __init__(
        self,
        driver: neo4j.Driver,
        filter_query: Optional[str] = None,
        resolve_properties: Optional[List[str]] = None,
        similarity_threshold: float = 0.8,
        neo4j_database: Optional[str] = None,
    ) -> None:
        if not IS_RAPIDFUZZ_INSTALLED:
            raise ImportError("""`rapidfuzz` python module needs to be installed to use
            the SpaCySemanticMatchResolver. Install it with:
            `pip install "neo4j-graphrag[fuzzy-matching]"`
            """)
        super().__init__(
            driver,
            filter_query,
            resolve_properties,
            similarity_threshold,
            neo4j_database,
        )

    async def run(self) -> ResolutionStats:
        return await super().run()

    def compute_similarity(self, text_a: str, text_b: str) -> float:
        # RapidFuzz's fuzz.WRatio returns a score from 0 to 100
        # normalize the input strings before the comparison is done (processor=utils.default_process)
        # e.g., lowercase the text, strip whitespace, and remove punctuation
        # normalize the score to the 0..1 range
        return fuzz.WRatio(text_a, text_b, processor=utils.default_process) / 100.0



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from __future__ import annotations

from unittest.mock import MagicMock, patch

import neo4j
import pytest
from neo4j.exceptions import CypherSyntaxError
from neo4j_graphrag.exceptions import (
    EmbeddingRequiredError,
    RetrieverInitializationError,
    SearchValidationError,
)
from neo4j_graphrag.neo4j_queries import get_search_query
from neo4j_graphrag.retrievers import VectorCypherRetriever, VectorRetriever
from neo4j_graphrag.types import (
    RetrieverResult,
    RetrieverResultItem,
    SearchType,
)


def test_vector_retriever_initialization(driver: MagicMock) -> None:
    with patch("neo4j_graphrag.retrievers.base.get_version") as mock_get_version:
        mock_get_version.return_value = ((5, 23, 0), False, False)
        VectorRetriever(driver=driver, index_name="my-index")
        mock_get_version.assert_called_once()


@patch("neo4j_graphrag.retrievers.base.get_version")
def test_vector_retriever_invalid_index_name(
    mock_get_version: MagicMock, driver: MagicMock
) -> None:
    mock_get_version.return_value = ((5, 23, 0), False, False)
    with pytest.raises(RetrieverInitializationError) as exc_info:
        VectorRetriever(driver=driver, index_name=42)  # type: ignore

    assert "index_name" in str(exc_info.value)
    assert "Input should be a valid string" in str(exc_info.value)


@patch("neo4j_graphrag.retrievers.base.get_version")
def test_vector_retriever_invalid_database_name(
    mock_get_version: MagicMock, driver: MagicMock
) -> None:
    mock_get_version.return_value = ((5, 23, 0), False, False)
    with pytest.raises(RetrieverInitializationError) as exc_info:
        VectorRetriever(
            driver=driver,
            index_name="my-index",
            neo4j_database=42,  # type: ignore
        )

    assert "neo4j_database" in str(exc_info.value)
    assert "Input should be a valid string" in str(exc_info.value)


@patch("neo4j_graphrag.retrievers.base.get_version")
def test_vector_cypher_retriever_invalid_retrieval_query(
    mock_get_version: MagicMock, driver: MagicMock
) -> None:
    mock_get_version.return_value = ((5, 23, 0), False, False)
    with pytest.raises(RetrieverInitializationError) as exc_info:
        VectorCypherRetriever(driver=driver, index_name="my-index", retrieval_query=42)  # type: ignore

        assert "retrieval_query" in str(exc_info.value)
        assert "Input should be a valid string" in str(exc_info.value)


@patch("neo4j_graphrag.retrievers.base.get_version")
def test_vector_cypher_retriever_invalid_database_name(
    mock_get_version: MagicMock, driver: MagicMock
) -> None:
    mock_get_version.return_value = ((5, 23, 0), False, False)
    retrieval_query = """
        RETURN node.id AS node_id, node.text AS text, score
        """
    with pytest.raises(RetrieverInitializationError) as exc_info:
        VectorCypherRetriever(
            driver=driver,
            index_name="my-index",
            retrieval_query=retrieval_query,
            neo4j_database=42,  # type: ignore
        )

        assert "neo4j_database" in str(exc_info.value)
        assert "Input should be a valid string" in str(exc_info.value)


def test_vector_cypher_retriever_initialization(driver: MagicMock) -> None:
    with patch("neo4j_graphrag.retrievers.base.get_version") as mock_get_version:
        mock_get_version.return_value = ((5, 23, 0), False, False)
        VectorCypherRetriever(driver=driver, index_name="my-index", retrieval_query="")
        mock_get_version.assert_called_once()


@patch("neo4j_graphrag.retrievers.VectorRetriever._fetch_index_infos")
@patch("neo4j_graphrag.retrievers.base.get_version")
def test_similarity_search_vector_happy_path(
    mock_get_version: MagicMock,
    _fetch_index_infos: MagicMock,
    driver: MagicMock,
    neo4j_record: MagicMock,
) -> None:
    mock_get_version.return_value = ((5, 23, 0), False, False)
    index_name = "my-index"
    dimensions = 1536
    query_vector = [1.0 for _ in range(dimensions)]
    top_k = 5
    effective_search_ratio = 2
    database = "neo4j"
    retriever = VectorRetriever(driver, index_name, neo4j_database=database)
    expected_records = [neo4j.Record({"node": {"text": "dummy-node"}, "score": 1.0})]
    retriever.driver.execute_query.return_value = [  # type: ignore
        expected_records,
        None,
        None,
    ]
    search_query, _ = get_search_query(SearchType.VECTOR)

    records = retriever.search(
        query_vector=query_vector,
        top_k=top_k,
        effective_search_ratio=effective_search_ratio,
    )

    retriever.driver.execute_query.assert_called_once_with(  # type: ignore
        search_query,
        {
            "vector_index_name": index_name,
            "top_k": top_k,
            "effective_search_ratio": effective_search_ratio,
            "query_vector": query_vector,
        },
        database_=database,
        routing_=neo4j.RoutingControl.READ,
    )
    assert records == RetrieverResult(
        items=[
            RetrieverResultItem(
                content="{'text': 'dummy-node'}",
                metadata={"score": 1.0, "nodeLabels": None, "id": None},
            ),
        ],
        metadata={"__retriever": "VectorRetriever", "query_vector": query_vector},
    )


@patch("neo4j_graphrag.retrievers.VectorRetriever._fetch_index_infos")
@patch("neo4j_graphrag.retrievers.base.get_version")
def test_similarity_search_text_happy_path(
    mock_get_version: MagicMock,
    _fetch_index_infos: MagicMock,
    driver: MagicMock,
    embedder: MagicMock,
    neo4j_record: MagicMock,
) -> None:
    mock_get_version.return_value = ((5, 23, 0), False, False)
    embed_query_vector = [1.0 for _ in range(1536)]
    embedder.embed_query.return_value = embed_query_vector
    index_name = "my-index"
    query_text = "may thy knife chip and shatter"
    top_k = 5
    effective_search_ratio = 2
    retriever = VectorRetriever(driver, index_name, embedder)
    driver.execute_query.return_value = [
        [neo4j_record],
        None,
        None,
    ]
    search_query, _ = get_search_query(SearchType.VECTOR)

    records = retriever.search(
        query_text=query_text,
        top_k=top_k,
        effective_search_ratio=effective_search_ratio,
    )

    embedder.embed_query.assert_called_once_with(query_text)
    driver.execute_query.assert_called_once_with(
        search_query,
        {
            "vector_index_name": index_name,
            "top_k": top_k,
            "effective_search_ratio": effective_search_ratio,
            "query_vector": embed_query_vector,
        },
        database_=None,
        routing_=neo4j.RoutingControl.READ,
    )
    assert records == RetrieverResult(
        items=[
            RetrieverResultItem(
                content="dummy-node",
                metadata={"score": 1.0, "nodeLabels": None, "id": None},
            ),
        ],
        metadata={"__retriever": "VectorRetriever", "query_vector": embed_query_vector},
    )


@patch("neo4j_graphrag.retrievers.VectorRetriever._fetch_index_infos")
@patch("neo4j_graphrag.retrievers.base.get_version")
def test_similarity_search_text_return_properties(
    mock_get_version: MagicMock,
    _fetch_index_infos: MagicMock,
    driver: MagicMock,
    embedder: MagicMock,
    neo4j_record: MagicMock,
) -> None:
    mock_get_version.return_value = ((5, 23, 0), False, False)
    embed_query_vector = [1.0 for _ in range(3)]
    embedder.embed_query.return_value = embed_query_vector
    index_name = "my-index"
    query_text = "may thy knife chip and shatter"
    top_k = 5
    effective_search_ratio = 2
    return_properties = ["node-property-1", "node-property-2"]

    retriever = VectorRetriever(
        driver, index_name, embedder, return_properties=return_properties
    )

    driver.execute_query.return_value = [
        [
            neo4j_record,
        ],
        None,
        None,
    ]
    search_query, _ = get_search_query(
        search_type=SearchType.VECTOR, return_properties=return_properties
    )

    records = retriever.search(
        query_text=query_text,
        top_k=top_k,
        effective_search_ratio=effective_search_ratio,
    )

    embedder.embed_query.assert_called_once_with(query_text)
    driver.execute_query.assert_called_once_with(
        search_query.rstrip(),
        {
            "vector_index_name": index_name,
            "top_k": top_k,
            "effective_search_ratio": effective_search_ratio,
            "query_vector": embed_query_vector,
        },
        database_=None,
        routing_=neo4j.RoutingControl.READ,
    )
    assert records == RetrieverResult(
        items=[
            RetrieverResultItem(
                content="dummy-node",
                metadata={"score": 1.0, "nodeLabels": None, "id": None},
            ),
        ],
        metadata={"__retriever": "VectorRetriever", "query_vector": embed_query_vector},
    )


def test_vector_retriever_search_missing_embedder_for_text(
    vector_retriever: VectorRetriever,
) -> None:
    query_text = "may thy knife chip and shatter"
    top_k = 5

    with pytest.raises(
        EmbeddingRequiredError, match="Embedding method required for text query"
    ):
        vector_retriever.search(query_text=query_text, top_k=top_k)


def test_vector_retriever_search_both_text_and_vector(
    vector_retriever: VectorRetriever,
) -> None:
    query_text = "may thy knife chip and shatter"
    query_vector = [1.1, 2.2, 3.3]
    top_k = 5

    with pytest.raises(
        SearchValidationError,
        match="You must provide exactly one of query_vector or query_text.",
    ):
        vector_retriever.search(
            query_text=query_text,
            query_vector=query_vector,
            top_k=top_k,
        )


@patch("neo4j_graphrag.retrievers.VectorRetriever._fetch_index_infos")
@patch("neo4j_graphrag.retrievers.base.get_version")
def test_vector_retriever_with_result_format_function(
    mock_get_version: MagicMock,
    _fetch_index_infos: MagicMock,
    driver: MagicMock,
    embedder: MagicMock,
    neo4j_record: MagicMock,
    result_formatter: MagicMock,
) -> None:
    mock_get_version.return_value = ((5, 23, 0), False, False)
    embed_query_vector = [1.0 for _ in range(1536)]
    embedder.embed_query.return_value = embed_query_vector
    index_name = "my-index"

    retriever = VectorRetriever(
        driver,
        index_name,
        embedder=embedder,
        result_formatter=result_formatter,
    )
    query_text = "may thy knife chip and shatter"
    top_k = 5
    driver.execute_query.return_value = [
        [neo4j_record],
        None,
        None,
    ]

    records = retriever.search(
        query_text=query_text,
        top_k=top_k,
    )

    assert records == RetrieverResult(
        items=[
            RetrieverResultItem(
                content="dummy-node", metadata={"score": 1.0, "node_id": 123}
            ),
        ],
        metadata={"__retriever": "VectorRetriever", "query_vector": embed_query_vector},
    )


def test_vector_cypher_retriever_search_missing_embedder_for_text(
    vector_cypher_retriever: VectorCypherRetriever,
) -> None:
    query_text = "may thy knife chip and shatter"
    top_k = 5

    with pytest.raises(
        EmbeddingRequiredError, match="Embedding method required for text query"
    ):
        vector_cypher_retriever.search(query_text=query_text, top_k=top_k)


def test_vector_cypher_retriever_search_both_text_and_vector(
    vector_cypher_retriever: VectorCypherRetriever,
) -> None:
    query_text = "may thy knife chip and shatter"
    query_vector = [1.1, 2.2, 3.3]
    top_k = 5

    with pytest.raises(
        SearchValidationError,
        match="You must provide exactly one of query_vector or query_text.",
    ):
        vector_cypher_retriever.search(
            query_text=query_text,
            query_vector=query_vector,
            top_k=top_k,
        )


@patch("neo4j_graphrag.retrievers.VectorCypherRetriever._fetch_index_infos")
@patch("neo4j_graphrag.retrievers.base.get_version")
def test_retrieval_query_happy_path(
    mock_get_version: MagicMock,
    _fetch_index_infos: MagicMock,
    driver: MagicMock,
    embedder: MagicMock,
) -> None:
    mock_get_version.return_value = ((5, 23, 0), False, False)
    embed_query_vector = [1.0 for _ in range(1536)]
    embedder.embed_query.return_value = embed_query_vector
    index_name = "my-index"
    retrieval_query = """
        RETURN node.id AS node_id, node.text AS text, score
        """
    database = "neo4j"
    retriever = VectorCypherRetriever(
        driver,
        index_name,
        retrieval_query,
        embedder=embedder,
        neo4j_database=database,
    )
    query_text = "may thy knife chip and shatter"
    top_k = 5
    effective_search_ratio = 2
    record = neo4j.Record({"node_id": 123, "text": "dummy-text", "score": 1.0})
    driver.execute_query.return_value = [
        [record],
        None,
        None,
    ]
    search_query, _ = get_search_query(
        SearchType.VECTOR, retrieval_query=retrieval_query
    )

    records = retriever.search(
        query_text=query_text,
        top_k=top_k,
        effective_search_ratio=effective_search_ratio,
    )

    embedder.embed_query.assert_called_once_with(query_text)
    driver.execute_query.assert_called_once_with(
        search_query,
        {
            "vector_index_name": index_name,
            "top_k": top_k,
            "effective_search_ratio": effective_search_ratio,
            "query_vector": embed_query_vector,
        },
        database_=database,
        routing_=neo4j.RoutingControl.READ,
    )
    assert records == RetrieverResult(
        items=[
            RetrieverResultItem(
                content="<Record node_id=123 text='dummy-text' score=1.0>",
                metadata=None,
            ),
        ],
        metadata={
            "__retriever": "VectorCypherRetriever",
            "query_vector": embed_query_vector,
        },
    )


@patch("neo4j_graphrag.retrievers.VectorCypherRetriever._fetch_index_infos")
@patch("neo4j_graphrag.retrievers.base.get_version")
def test_retrieval_query_with_result_format_function(
    mock_get_version: MagicMock,
    _fetch_index_infos: MagicMock,
    driver: MagicMock,
    embedder: MagicMock,
    neo4j_record: MagicMock,
    result_formatter: MagicMock,
) -> None:
    mock_get_version.return_value = ((5, 23, 0), False, False)
    embed_query_vector = [1.0 for _ in range(1536)]
    embedder.embed_query.return_value = embed_query_vector
    index_name = "my-index"
    retrieval_query = """
        RETURN node.id AS node_id, node.text AS text, score
        """

    retriever = VectorCypherRetriever(
        driver,
        index_name,
        retrieval_query,
        embedder=embedder,
        result_formatter=result_formatter,
    )
    query_text = "may thy knife chip and shatter"
    top_k = 5
    effective_search_ratio = 2
    driver.execute_query.return_value = [
        [neo4j_record],
        None,
        None,
    ]
    search_query, _ = get_search_query(
        SearchType.VECTOR, retrieval_query=retrieval_query
    )

    records = retriever.search(
        query_text=query_text,
        top_k=top_k,
        effective_search_ratio=effective_search_ratio,
    )

    embedder.embed_query.assert_called_once_with(query_text)
    driver.execute_query.assert_called_once_with(
        search_query,
        {
            "vector_index_name": index_name,
            "top_k": top_k,
            "effective_search_ratio": effective_search_ratio,
            "query_vector": embed_query_vector,
        },
        database_=None,
        routing_=neo4j.RoutingControl.READ,
    )
    assert records == RetrieverResult(
        items=[
            RetrieverResultItem(
                content="dummy-node", metadata={"score": 1.0, "node_id": 123}
            ),
        ],
        metadata={
            "__retriever": "VectorCypherRetriever",
            "query_vector": embed_query_vector,
        },
    )


@patch("neo4j_graphrag.retrievers.VectorCypherRetriever._fetch_index_infos")
@patch("neo4j_graphrag.retrievers.base.get_version")
def test_retrieval_query_with_params(
    mock_get_version: MagicMock,
    _fetch_index_infos: MagicMock,
    driver: MagicMock,
    embedder: MagicMock,
) -> None:
    mock_get_version.return_value = ((5, 23, 0), False, False)
    embed_query_vector = [1.0 for _ in range(1536)]
    embedder.embed_query.return_value = embed_query_vector
    index_name = "my-index"
    retrieval_query = """
        RETURN node.id AS node_id, node.text AS text, score, {test: $param} AS metadata
        """
    query_params = {
        "param": "dummy-param",
    }
    retriever = VectorCypherRetriever(
        driver,
        index_name,
        retrieval_query,
        embedder=embedder,
    )
    query_text = "may thy knife chip and shatter"
    top_k = 5
    effective_search_ratio = 2
    driver.execute_query.return_value = [
        [neo4j.Record({"node_id": 123, "text": "dummy-text", "score": 1.0})],
        None,
        None,
    ]
    search_query, _ = get_search_query(
        SearchType.VECTOR, retrieval_query=retrieval_query
    )

    records = retriever.search(
        query_text=query_text,
        top_k=top_k,
        effective_search_ratio=effective_search_ratio,
        query_params=query_params,
    )

    embedder.embed_query.assert_called_once_with(query_text)

    driver.execute_query.assert_called_once_with(
        search_query,
        {
            "vector_index_name": index_name,
            "top_k": top_k,
            "effective_search_ratio": effective_search_ratio,
            "query_vector": embed_query_vector,
            "param": "dummy-param",
        },
        database_=None,
        routing_=neo4j.RoutingControl.READ,
    )

    assert records == RetrieverResult(
        items=[
            RetrieverResultItem(
                content="<Record node_id=123 text='dummy-text' score=1.0>",
                metadata=None,
            ),
        ],
        metadata={
            "__retriever": "VectorCypherRetriever",
            "query_vector": embed_query_vector,
        },
    )


@patch("neo4j_graphrag.retrievers.VectorCypherRetriever._fetch_index_infos")
@patch("neo4j_graphrag.retrievers.base.get_version")
def test_retrieval_query_cypher_error(
    mock_get_version: MagicMock,
    _fetch_index_infos: MagicMock,
    driver: MagicMock,
    embedder: MagicMock,
) -> None:
    mock_get_version.return_value = ((5, 23, 0), False, False)
    embed_query_vector = [1.0 for _ in range(1536)]
    embedder.embed_query.return_value = embed_query_vector
    index_name = "my-index"
    retrieval_query = """
        this is not a cypher query
        """
    retriever = VectorCypherRetriever(
        driver, index_name, retrieval_query, embedder=embedder
    )
    query_text = "may thy knife chip and shatter"
    top_k = 5
    driver.execute_query.side_effect = CypherSyntaxError

    with pytest.raises(CypherSyntaxError):
        retriever.search(
            query_text=query_text,
            top_k=top_k,
        )



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from __future__ import annotations

from typing import Any, List, Optional, Sequence

import neo4j

from neo4j_graphrag.llm.base import LLMInterface
from neo4j_graphrag.retrievers.base import Retriever
from neo4j_graphrag.types import RawSearchResult
from neo4j_graphrag.tool import Tool
from neo4j_graphrag.types import LLMMessage


class ToolsRetriever(Retriever):
    """A retriever that uses an LLM to select appropriate tools for retrieval based on user input.

    This retriever takes an LLM instance and a list of Tool objects as input. When a search is performed,
    it uses the LLM to analyze the query and determine which tools (if any) should be used to retrieve
    the necessary data. It then executes the selected tools and returns the combined results.

    Example:

    .. code-block:: python

        import neo4j
        from neo4j_graphrag.retrievers import ToolsRetriever, VectorRetriever, Text2CypherRetriever
        from neo4j_graphrag.llm import OpenAILLM
        from neo4j_graphrag.embeddings import OpenAIEmbeddings

        driver = neo4j.GraphDatabase.driver("neo4j://localhost:7687", auth=("neo4j", "password"))
        llm = OpenAILLM(model_name="gpt-4", api_key="your-api-key")
        embedder = OpenAIEmbeddings(model="text-embedding-3-small", api_key="your-api-key")

        # Create retrievers and convert them to tools
        vector_retriever = VectorRetriever(driver, "vector-index", embedder)
        vector_tool = vector_retriever.convert_to_tool(
            name="vector_search",
            description="Search for documents using semantic similarity"
        )

        text2cypher_retriever = Text2CypherRetriever(driver, llm)
        cypher_tool = text2cypher_retriever.convert_to_tool(
            name="cypher_search",
            description="Generate and execute Cypher queries for structured data retrieval"
        )

        # Initialize ToolsRetriever with the tools
        tools_retriever = ToolsRetriever(
            driver=driver,
            llm=llm,
            tools=[vector_tool, cypher_tool]
        )

        # Use the retriever - the LLM will automatically select appropriate tools
        result = tools_retriever.search("What movies did Tom Hanks act in and what are their plots?")

    Args:
        driver (neo4j.Driver): The Neo4j Python driver.
        llm (LLMInterface): LLM instance used to select and coordinate tool execution.
        tools (Sequence[Tool]): List of tools available for selection. All tools must have unique names.
        neo4j_database (Optional[str]): The name of the Neo4j database. If not provided, this defaults to the server's default database ("neo4j" by default).
        system_instruction (Optional[str]): Custom system instruction for the LLM to guide tool selection. If not provided, a default instruction is used.

    Raises:
        ValueError: If duplicate tool names are found in the tools list.
    """

    # Disable Neo4j version verification since this retriever doesn't directly interact with Neo4j
    VERIFY_NEO4J_VERSION = False

    def __init__(
        self,
        driver: neo4j.Driver,
        llm: LLMInterface,
        tools: Sequence[Tool],
        neo4j_database: Optional[str] = None,
        system_instruction: Optional[str] = None,
    ):
        """Initialize the ToolsRetriever with an LLM and a list of tools."""
        super().__init__(driver, neo4j_database)
        self.llm = llm
        self._tools = list(tools)  # Make a copy to allow modification
        self._validate_tool_names()
        self.system_instruction = (
            system_instruction or self._get_default_system_instruction()
        )

    def _validate_tool_names(self) -> None:
        """Validate that all tool names are unique."""
        tool_names = [tool.get_name() for tool in self._tools]
        duplicate_names = [
            name for name in set(tool_names) if tool_names.count(name) > 1
        ]

        if duplicate_names:
            raise ValueError(
                f"Duplicate tool names found: {duplicate_names}. "
                "All tools must have unique names for proper LLM tool selection."
            )

    def _get_default_system_instruction(self) -> str:
        """Get the default system instruction for the LLM."""
        return (
            "You are an assistant that helps select the most appropriate tools to retrieve information "
            "based on the user's query. Analyze the query carefully and determine which tools, if any, "
            "would be most helpful in retrieving the relevant information. You can select multiple tools "
            "if necessary, or none if no tools are appropriate for the query."
        )

    def get_search_results(
        self,
        query_text: str,
        message_history: Optional[List[LLMMessage]] = None,
        **kwargs: Any,
    ) -> RawSearchResult:
        """Use the LLM to select and execute appropriate tools based on the query.

        Args:
            query_text (str): The user's query text.
            message_history (Optional[Union[List[LLMMessage], MessageHistory]], optional):
                Previous conversation history. Defaults to None.
            **kwargs (Any): Additional arguments passed to the tool execution.

        Returns:
            RawSearchResult: The combined results from the executed tools.
        """
        if not self._tools:
            # No tools available, return empty result
            return RawSearchResult(
                records=[],
                metadata={"query": query_text, "error": "No tools available"},
            )

        try:
            # Use the LLM to select appropriate tools
            tool_call_response = self.llm.invoke_with_tools(
                input=query_text,
                tools=self._tools,
                message_history=message_history,
                system_instruction=self.system_instruction,
            )
            # If no tool calls were made, return empty result
            if not tool_call_response.tool_calls:
                return RawSearchResult(
                    records=[],
                    metadata={
                        "query": query_text,
                        "llm_response": tool_call_response.content,
                        "tools_selected": [],
                    },
                )

            # Execute each selected tool and collect results
            all_records = []
            tools_selected = []

            for tool_call in tool_call_response.tool_calls:
                tool_name = tool_call.name
                tools_selected.append(tool_name)

                # Find the tool by name
                selected_tool = next(
                    (tool for tool in self._tools if tool.get_name() == tool_name), None
                )
                if selected_tool is not None:
                    # Extract arguments from the tool call
                    tool_args = tool_call.arguments or {}

                    # Execute the tool with the provided arguments
                    tool_result = selected_tool.execute(**tool_args)

                    # Handle different tool result types
                    if hasattr(tool_result, "items") and not callable(
                        getattr(tool_result, "items")
                    ):
                        # RetrieverResult from formatted retriever tools
                        for item in tool_result.items:
                            record = neo4j.Record(
                                {
                                    "content": item.content,
                                    "tool_name": tool_name,
                                    "metadata": {
                                        **(item.metadata or {}),
                                        "tool": tool_name,
                                    },
                                }
                            )
                            all_records.append(record)
                    elif hasattr(tool_result, "records"):
                        # RawSearchResult from raw retriever tools (legacy)
                        for record in tool_result.records:
                            # Wrap raw records with tool attribution
                            attributed_record = neo4j.Record(
                                {
                                    "content": str(record),
                                    "tool_name": tool_name,
                                    "metadata": {
                                        "original_record": dict(record),
                                        "tool": tool_name,
                                    },
                                }
                            )
                            all_records.append(attributed_record)
                    else:
                        # Handle non-retriever tools or simple return values
                        record = neo4j.Record(
                            {
                                "content": str(tool_result),
                                "tool_name": tool_name,
                                "metadata": {"tool": tool_name},
                            }
                        )
                        all_records.append(record)

            # Combine metadata from all tool calls
            combined_metadata = {
                "query": query_text,
                "llm_response": tool_call_response.content,
                "tools_selected": tools_selected,
            }

            return RawSearchResult(records=all_records, metadata=combined_metadata)

        except Exception as e:
            # Handle any errors during tool selection or execution
            return RawSearchResult(
                records=[],
                metadata={
                    "query": query_text,
                    "error": str(e),
                    "error_type": type(e).__name__,
                },
            )



from neo4j_graphrag.llm import LLMResponse, OpenAILLM

# set api key here on in the OPENAI_API_KEY env var
api_key = None

llm = OpenAILLM(model_name="gpt-4o", api_key=api_key)
res: LLMResponse = llm.invoke("say something")
print(res.content)



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import neo4j
import pytest
from neo4j_graphrag.experimental.components.resolver import (
    SinglePropertyExactMatchResolver,
)


@pytest.mark.asyncio
@pytest.mark.usefixtures("setup_neo4j_for_kg_construction")
async def test_resolver_single_node(driver: neo4j.Driver) -> None:
    driver.execute_query("MATCH (n) DETACH DELETE n")
    driver.execute_query(
        """
        CREATE (d:Document {id: "0", path: "path"})
        CREATE (c:Chunk {id: "0:0"})
        CREATE (c)-[:FROM_DOCUMENT]->(d)
        CREATE (alice:__Entity__:Person {id: "0:0:1", name: "Alice"})
        CREATE (alice)-[:FROM_CHUNK]->(c)
        """
    )
    resolver = SinglePropertyExactMatchResolver(driver)
    res = await resolver.run()
    # __Entity__ nodes attached to a chunk
    assert res.number_of_nodes_to_resolve == 1
    # Alice
    assert res.number_of_created_nodes == 1

    records, _, _ = driver.execute_query(
        "MATCH path=(:Person {name: 'Alice'}) RETURN path"
    )
    assert len(records) == 1
    path = records[0].get("path")
    assert path.start_node.get("name") == "Alice"
    assert path.start_node.labels == frozenset({"__Entity__", "Person"})
    assert path.start_node.get("id") == "0:0:1"


@pytest.mark.asyncio
@pytest.mark.usefixtures("setup_neo4j_for_kg_construction")
async def test_resolver_two_nodes_and_relationships(driver: neo4j.Driver) -> None:
    driver.execute_query("MATCH (n) DETACH DELETE n")
    driver.execute_query(
        """
        CREATE (d:Document {id: "0", path: "path"})
        CREATE (c:Chunk {id: "0:0"})
        CREATE (c)-[:FROM_DOCUMENT]->(d)
        CREATE (alice1:__Entity__:Person {id: "0:0:1", name: "Alice"})
        CREATE (alice2:__Entity__:Person {id: "0:0:2", name: "Alice"})
        CREATE (sweden:__Entity__:Country {id: "0:0:3", name: "Sweden"})
        CREATE (alice1)-[:LIVES_IN]->(sweden)
        CREATE (alice1)-[:FROM_CHUNK]->(c)
        CREATE (alice2)-[:FROM_CHUNK]->(c)
        CREATE (sweden)-[:FROM_CHUNK]->(c)
        """
    )
    resolver = SinglePropertyExactMatchResolver(driver)
    res = await resolver.run()
    # __Entity__ nodes attached to a chunk
    assert res.number_of_nodes_to_resolve == 3
    # Alice and Sweden
    assert res.number_of_created_nodes == 2

    # check the domain graph
    records, _, _ = driver.execute_query(
        "MATCH path=(:Person {name: 'Alice'})-[:LIVES_IN]->(:Country {name: 'Sweden'}) RETURN path"
    )
    assert len(records) == 1
    path = records[0].get("path")
    assert path.start_node.get("name") == "Alice"
    assert path.start_node.labels == frozenset({"__Entity__", "Person"})
    assert path.end_node.get("name") == "Sweden"
    assert path.end_node.labels == frozenset({"__Entity__", "Country"})
    assert len(path.relationships) == 1
    assert path.relationships[0].type == "LIVES_IN"

    # check the lexical graph
    records, _, _ = driver.execute_query(
        "MATCH path=(:Person {name: 'Alice'})-[:FROM_CHUNK]->(:Chunk) RETURN path"
    )
    assert len(records) == 1
    path = records[0].get("path")
    assert path.start_node.get("name") == "Alice"
    assert path.end_node.labels == frozenset({"Chunk"})
    assert path.end_node.get("id") == "0:0"


@pytest.mark.asyncio
@pytest.mark.usefixtures("setup_neo4j_for_kg_construction")
async def test_resolver_same_name_different_labels(driver: neo4j.Driver) -> None:
    driver.execute_query("MATCH (n) DETACH DELETE n")
    driver.execute_query(
        """
        CREATE (d:Document {id: "0", path: "path"})
        CREATE (c:Chunk {id: "0:0"})
        CREATE (c)-[:FROM_DOCUMENT]->(d)
        CREATE (alice1:__Entity__:Person {id: "0:0:1", name: "Alice"})
        CREATE (alice2:__Entity__:Human {id: "0:0:2", name: "Alice"})
        CREATE (alice1)-[:FROM_CHUNK]->(c)
        CREATE (alice2)-[:FROM_CHUNK]->(c)
        CREATE (sweden)-[:FROM_CHUNK]->(c)
        """
    )
    resolver = SinglePropertyExactMatchResolver(driver)
    res = await resolver.run()
    # __Entity__ nodes attached to a chunk
    assert res.number_of_nodes_to_resolve == 2
    # Alice Person and Alice Human
    assert res.number_of_created_nodes == 2

    records, _, _ = driver.execute_query("MATCH (alice {name: 'Alice'}) RETURN alice")
    assert len(records) == 2


@pytest.mark.asyncio
@pytest.mark.usefixtures("setup_neo4j_for_kg_construction")
async def test_resolver_custom_property(driver: neo4j.Driver) -> None:
    driver.execute_query("MATCH (n) DETACH DELETE n")
    driver.execute_query(
        """
        CREATE (d:Document {id: "0", path: "path"})
        CREATE (c:Chunk {id: "0:0"})
        CREATE (c)-[:FROM_DOCUMENT]->(d)
        CREATE (alice:__Entity__:Person {id: "0:0:1", name: "Alice"})
        CREATE (alicia:__Entity__:Person {id: "0:0:1", name: "Alicia"})
        CREATE (alice)-[:FROM_CHUNK]->(c)
        CREATE (alicia)-[:FROM_CHUNK]->(c)
        """
    )
    resolver = SinglePropertyExactMatchResolver(driver, resolve_property="id")
    res = await resolver.run()
    # __Entity__ nodes attached to a chunk
    assert res.number_of_nodes_to_resolve == 2
    assert res.number_of_created_nodes == 1

    records, _, _ = driver.execute_query("MATCH (person:Person) RETURN person")
    assert len(records) == 1
    assert records[0].get("person").get("name") in ["Alice", "Alicia"]


@pytest.mark.asyncio
@pytest.mark.usefixtures("setup_neo4j_for_kg_construction")
async def test_resolver_custom_filter(driver: neo4j.Driver) -> None:
    driver.execute_query("MATCH (n) DETACH DELETE n")
    driver.execute_query(
        """
        CREATE (d:Document {id: "0", path: "path"})
        CREATE (c:Chunk {id: "0:0"})
        CREATE (c)-[:FROM_DOCUMENT]->(d)
        CREATE (alice1:__Entity__:Person {id: "0:0:1", name: "Alice"})
        CREATE (alice2:__Entity__:Person {id: "0:0:2", name: "Alice"})
        CREATE (sweden:__Entity__:Country {id: "0:0:3", name: "Sweden"})
        CREATE (alice1)-[:LIVES_IN]->(sweden)
        CREATE (alice1)-[:FROM_CHUNK]->(c)
        CREATE (alice2)-[:FROM_CHUNK]->(c)
        CREATE (sweden)-[:FROM_CHUNK]->(c)
        """
    )
    resolver = SinglePropertyExactMatchResolver(
        driver, filter_query="WHERE not entity:Person"
    )
    res = await resolver.run()
    # __Entity__ nodes attached to a chunk without a Person label
    # so only Country here
    assert res.number_of_nodes_to_resolve == 1
    # Sweden
    assert res.number_of_created_nodes == 1



"""This example demonstrates how to customize the retriever
results format.

Retriever.get_search_result returns a RawSearchResult object that consists
in a list of neo4j.Records and an optional metadata dictionary. The
Retriever.search method returns a RetrieverResult object where each neo4j.Record
has been replaced by a RetrieverResultItem, ie a content and metadata dictionary.
The content is what will be used to augment the prompt. By default, this content
is a stringified representation of the neo4j.Record. There are multiple ways the
user can act on this format:
- Use the `return_properties` parameter
- And/or use the result_formatter parameter

Let's consider the movie database where the movies' plot have been embedded. Movie
nodes have additional properties such as title and are connected to Actor nodes that
have a name property:

(:Movie {embedding: [], title: "", plot: "", year: "", budget: 1000, ....})
    <-[:ACTED_IN]-
    (:Actor {name: ...})

NB: to run this example OPENAI_API_KEY needs to be in the env vars.
To use another embedder, see the corresponding examples in ../customize/embeddings
"""

import neo4j
from neo4j_graphrag.embeddings.openai import OpenAIEmbeddings
from neo4j_graphrag.retrievers import VectorRetriever
from neo4j_graphrag.types import RetrieverResultItem

URI = "neo4j+s://demo.neo4jlabs.com"
AUTH = ("recommendations", "recommendations")
DATABASE = "recommendations"
INDEX_NAME = "moviePlotsEmbedding"


# Connect to Neo4j database
driver = neo4j.GraphDatabase.driver(URI, auth=AUTH)


query_text = "Find a movie about astronauts"
top_k_results = 1


"""First, let's select the properties we want to return
with the return_properties parameter:
"""
print("=" * 50)
print("RETURN PROPERTIES")
retriever = VectorRetriever(
    driver=driver,
    index_name=INDEX_NAME,
    embedder=OpenAIEmbeddings(),
    return_properties=["title", "plot"],
    neo4j_database=DATABASE,
)
print(retriever.search(query_text=query_text, top_k=top_k_results))
print()
"""
OUTPUT:
RetrieverResult(
    items=[
        RetrieverResultItem(
            content="{'title': 'For All Mankind', 'plot': 'This movie documents the Apollo missions perhaps the most definitively of any movie under two hours. Al Reinert watched all the footage shot during the missions--over 6,000,000 feet of it, ...'}",
            metadata={'score': 0.9354040622711182, 'nodeLabels': None, 'id': None}
        )
    ]
    metadata={'__retriever': 'VectorRetriever'}
)
"""


"""In a second example, we'll use the ability to format the result in detail
with a function
"""
print("=" * 50)
print("RESULT FOMATTER")


def my_result_formatter(record: neo4j.Record) -> RetrieverResultItem:
    """
    If 'return_properties' are not specified, vector retrievers will return a record with keys:
    - `node`: a dict representation of all node properties (except embedding if we can identify it)
    - `id`: the node element ID
    - `nodeLabels`: the labels attached to the node
    - `score`: the score returned by the vector index search that tells us how close the node vector is from the query vector

    In the case of movies, we may want to keep in the content only the title and plot
    (passed to the LLM afterward) and keep in the metadata the score.
    This can be achieved with this function:
    """
    node = record.get("node")
    return RetrieverResultItem(
        content=f"{node.get('title')}: {node.get('plot')}",
        metadata={"score": record.get("score")},
    )


retriever = VectorRetriever(
    driver=driver,
    index_name=INDEX_NAME,
    embedder=OpenAIEmbeddings(),
    result_formatter=my_result_formatter,
)

query_text = "Find a movie about astronauts"
print(retriever.search(query_text=query_text, top_k=top_k_results))
print()
"""
OUTPUT:
RetrieverResult(
    items=[
        RetrieverResultItem(
            content='For All Mankind: This movie documents the Apollo missions perhaps the most definitively of any movie under two hours. Al Reinert watched all the footage shot during the missions--over 6,000,000 feet of it, ...',
            metadata={'score': 0.9354040622711182}
        )
    ]
    metadata={'__retriever': 'VectorRetriever'}
)
"""


"""We can mix both return_properties and result_formatter:
"""
print("=" * 50)
print("RETURN PROPERTIES + RESULT FOMATTER")

retriever = VectorRetriever(
    driver=driver,
    index_name=INDEX_NAME,
    embedder=OpenAIEmbeddings(),
    return_properties=["title", "plot"],
    result_formatter=my_result_formatter,
)

query_text = "Find a movie about astronauts"
print(retriever.search(query_text=query_text, top_k=top_k_results))
print()

"""
OUTPUT:
RetrieverResult(
    items=[
        RetrieverResultItem(
            content='For All Mankind: This movie documents the Apollo missions perhaps the most definitively of any movie under two hours. Al Reinert watched all the footage shot during the missions--over 6,000,000 feet of it, ...',
            metadata={'score': 0.9354040622711182})
    ]
    metadata={'__retriever': 'VectorRetriever'}
)
"""

driver.close()



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

# Standard library imports
from typing import Any, List, cast
from unittest.mock import MagicMock

import neo4j

# Local imports
from neo4j_graphrag.llm.base import LLMInterface
from neo4j_graphrag.llm.types import ToolCall, ToolCallResponse
from neo4j_graphrag.retrievers.tools_retriever import ToolsRetriever
from neo4j_graphrag.tool import Tool


# Mock dependencies
def create_mock_driver() -> neo4j.Driver:
    driver = MagicMock(spec=neo4j.Driver)
    # Create a mock result object with a records attribute
    mock_result = MagicMock()
    mock_result.records = [MagicMock()]
    driver.execute_query.return_value = mock_result
    return cast(neo4j.Driver, driver)


def create_mock_llm() -> Any:
    llm = MagicMock(spec=LLMInterface)
    return llm


def create_mock_tool(name: str = "MockTool", description: str = "A mock tool") -> Any:
    tool = MagicMock(spec=Tool)
    cast(Any, tool.get_name).return_value = name
    cast(Any, tool.get_description).return_value = description
    cast(Any, tool.get_parameters).return_value = {
        "type": "object",
        "properties": {
            "query": {
                "type": "string",
                "description": "The query to search for",
            }
        },
    }
    # Mock the execute method to return a dictionary with records and metadata
    cast(Any, tool.execute).return_value = {
        "records": [neo4j.Record({"result": f"Result from {name}"})],
        "metadata": {"source": name},
    }
    return tool


class TestToolsRetriever:
    """Test the ToolsRetriever class."""

    def test_initialization(self) -> None:
        """Test that the ToolsRetriever initializes correctly."""
        driver = create_mock_driver()
        llm = create_mock_llm()
        tools = [create_mock_tool("Tool1"), create_mock_tool("Tool2")]

        retriever = ToolsRetriever(driver=driver, llm=llm, tools=tools)

        assert retriever.llm == llm
        assert len(retriever._tools) == 2
        assert retriever._tools[0].get_name() == "Tool1"
        assert retriever._tools[1].get_name() == "Tool2"

    def test_get_search_results_no_tools(self) -> None:
        """Test that get_search_results returns an empty result when no tools are available."""
        driver = create_mock_driver()
        llm = create_mock_llm()
        tools: List[Tool] = []

        retriever = ToolsRetriever(driver=driver, llm=llm, tools=tools)
        result = retriever.get_search_results(query_text="Test query")

        assert result.records == []
        assert result.metadata is not None
        assert result.metadata.get("query") == "Test query"
        assert "error" in result.metadata
        assert result.metadata.get("error") == "No tools available"

    def test_get_search_results_no_tool_calls(self) -> None:
        """Test that get_search_results returns an empty result when the LLM doesn't select any tools."""
        driver = create_mock_driver()
        llm = create_mock_llm()
        tools = [create_mock_tool("Tool1"), create_mock_tool("Tool2")]

        # Mock the LLM to return a response with no tool calls
        cast(Any, llm.invoke_with_tools).return_value = ToolCallResponse(
            content="I don't need any tools for this query.",
            tool_calls=[],
        )

        retriever = ToolsRetriever(driver=driver, llm=llm, tools=tools)
        result = retriever.get_search_results(query_text="Test query")

        assert result.records == []
        assert result.metadata is not None
        assert result.metadata.get("query") == "Test query"
        assert (
            result.metadata.get("llm_response")
            == "I don't need any tools for this query."
        )
        assert result.metadata.get("tools_selected") == []

    def test_get_search_results_with_tool_calls(self) -> None:
        """Test that get_search_results correctly executes selected tools and returns their results."""
        driver = create_mock_driver()
        llm = create_mock_llm()
        tool1 = create_mock_tool("Tool1")
        tool2 = create_mock_tool("Tool2")
        tools = [tool1, tool2]

        # Mock the LLM to return a response with tool calls
        cast(Any, llm.invoke_with_tools).return_value = ToolCallResponse(
            content="I'll use Tool1 for this query.",
            tool_calls=[
                ToolCall(
                    name="Tool1",
                    arguments={"query": "Test query"},
                )
            ],
        )

        # Mock the tool execution to return a simple string value
        # This is processed by the ToolsRetriever and converted to a neo4j.Record
        cast(Any, tool1).execute.return_value = "Result from Tool1"

        retriever = ToolsRetriever(driver=driver, llm=llm, tools=tools)
        result = retriever.get_search_results(query_text="Test query")

        # Check that the LLM was called with the right arguments
        cast(Any, llm.invoke_with_tools).assert_called_once_with(
            input="Test query",
            tools=tools,
            message_history=None,
            system_instruction=retriever.system_instruction,
        )

        # Check that the tool was executed with the right arguments
        tool1.execute.assert_called_once_with(query="Test query")

        # Check that the result contains the expected records and metadata
        assert len(result.records) == 1
        # The record is a neo4j.Record object
        assert isinstance(result.records[0], neo4j.Record)
        # Access the result directly using index 0
        assert result.records[0][0] == "Result from Tool1"
        assert result.metadata is not None
        assert result.metadata.get("query") == "Test query"
        assert result.metadata.get("llm_response") == "I'll use Tool1 for this query."
        assert result.metadata.get("tools_selected") == ["Tool1"]

    def test_get_search_results_with_multiple_tool_calls(self) -> None:
        """Test that get_search_results correctly executes multiple selected tools and combines their results."""
        driver = create_mock_driver()
        llm = create_mock_llm()
        tool1 = create_mock_tool("Tool1")
        tool2 = create_mock_tool("Tool2")
        tools = [tool1, tool2]

        # Mock the LLM to return a response with multiple tool calls
        cast(Any, llm.invoke_with_tools).return_value = ToolCallResponse(
            content="I'll use both Tool1 and Tool2 for this query.",
            tool_calls=[
                ToolCall(
                    name="Tool1",
                    arguments={"query": "Test query part 1"},
                ),
                ToolCall(
                    name="Tool2",
                    arguments={"query": "Test query part 2"},
                ),
            ],
        )

        # Mock the tool executions to return specific records
        tool1_record = neo4j.Record({"result": "Result from Tool1"})
        cast(Any, tool1.execute).return_value = {
            "records": [tool1_record],
            "metadata": {"source": "Tool1"},
        }

        tool2_record = neo4j.Record({"result": "Result from Tool2"})
        cast(Any, tool2.execute).return_value = {
            "records": [tool2_record],
            "metadata": {"source": "Tool2"},
        }

        retriever = ToolsRetriever(driver=driver, llm=llm, tools=tools)
        result = retriever.get_search_results(query_text="Test query")

        # Check that both tools were executed with the right arguments
        cast(Any, tool1.execute).assert_called_once_with(query="Test query part 1")
        cast(Any, tool2.execute).assert_called_once_with(query="Test query part 2")

        # Check that the result contains the expected records and metadata
        assert len(result.records) == 2
        assert result.metadata is not None
        assert result.metadata.get("query") == "Test query"
        assert (
            result.metadata.get("llm_response")
            == "I'll use both Tool1 and Tool2 for this query."
        )
        assert result.metadata.get("tools_selected") == ["Tool1", "Tool2"]

    def test_get_search_results_with_error(self) -> None:
        """Test that get_search_results handles errors during tool execution."""
        driver = create_mock_driver()
        llm = create_mock_llm()
        tool = create_mock_tool("Tool1")
        tools = [tool]

        # Mock the LLM to raise an exception
        cast(Any, llm.invoke_with_tools).side_effect = Exception("LLM error")

        retriever = ToolsRetriever(driver=driver, llm=llm, tools=tools)
        result = retriever.get_search_results(query_text="Test query")

        # Check that the result contains the error information
        assert result.records == []
        assert result.metadata is not None
        assert result.metadata.get("query") == "Test query"
        assert result.metadata.get("error") == "LLM error"
        assert result.metadata.get("error_type") == "Exception"

    def test_custom_system_instruction(self) -> None:
        """Test that a custom system instruction is used when provided."""
        driver = create_mock_driver()
        llm = create_mock_llm()
        tools = [create_mock_tool("Tool1")]
        custom_instruction = "This is a custom system instruction."

        retriever = ToolsRetriever(
            driver=driver, llm=llm, tools=tools, system_instruction=custom_instruction
        )

        assert retriever.system_instruction == custom_instruction

        # Test that the custom instruction is passed to the LLM
        retriever.get_search_results(query_text="Test query")

        llm.invoke_with_tools.assert_called_once_with(
            input="Test query",
            tools=tools,
            message_history=None,
            system_instruction=custom_instruction,
        )



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import pytest
from neo4j_graphrag.experimental.pipeline.pipeline_graph import (
    PipelineEdge,
    PipelineGraph,
    PipelineNode,
)


def test_node_alone() -> None:
    n = PipelineNode(name="node", data={})
    assert n.is_root() is True
    assert n.is_leaf() is True


def test_node_not_root() -> None:
    n = PipelineNode(name="node", data={})
    n.parents = ["other_node"]
    assert n.is_root() is False
    assert n.is_leaf() is True


def test_node_not_leaf() -> None:
    n = PipelineNode(name="node", data={})
    n.children = ["other_node"]
    assert n.is_root() is True
    assert n.is_leaf() is False


def test_graph_add_nodes() -> None:
    g: PipelineGraph[PipelineNode, PipelineEdge] = PipelineGraph()
    n1 = PipelineNode("n1", {})
    n2 = PipelineNode("n2", {})
    g.add_node(n1)
    g.add_node(n2)
    assert len(g._nodes) == 2
    edge = PipelineEdge(n1.name, n2.name, {"key": "value"})
    g.add_edge(edge)
    assert len(g._edges) == 1

    assert n1.children == [n2.name]
    assert n2.parents == [n1.name]


@pytest.fixture(scope="function")
def graph() -> PipelineGraph[PipelineNode, PipelineEdge]:
    g: PipelineGraph[PipelineNode, PipelineEdge] = PipelineGraph()
    n1 = PipelineNode("n1", {})
    n2 = PipelineNode("n2", {})
    g.add_node(n1)
    g.add_node(n2)
    edge = PipelineEdge(n1.name, n2.name, {"key": "value"})
    g.add_edge(edge)
    return g


def test_graph_roots(graph: PipelineGraph[PipelineNode, PipelineEdge]) -> None:
    roots = graph.roots()
    assert len(roots) == 1
    assert roots[0].name == "n1"


def test_graph_next_edge(graph: PipelineGraph[PipelineNode, PipelineEdge]) -> None:
    start = graph._nodes["n1"]
    next_edges = graph.next_edges(start.name)
    assert len(next_edges) == 1
    next_edge = next_edges[0]
    assert isinstance(next_edge, PipelineEdge)
    assert next_edge.start == "n1"
    assert next_edge.end == "n2"


def test_graph_prev_edge(graph: PipelineGraph[PipelineNode, PipelineEdge]) -> None:
    start = graph._nodes["n2"]
    next_edges = graph.previous_edges(start.name)
    assert len(next_edges) == 1
    next_edge = next_edges[0]
    assert isinstance(next_edge, PipelineEdge)
    assert next_edge.start == "n1"
    assert next_edge.end == "n2"


def test_graph_contains(graph: PipelineGraph[PipelineNode, PipelineEdge]) -> None:
    start = graph._nodes["n2"]
    assert start in graph


def test_graph_is_cyclic(graph: PipelineGraph[PipelineNode, PipelineEdge]) -> None:
    g: PipelineGraph[PipelineNode, PipelineEdge] = PipelineGraph()
    n1 = PipelineNode("n1", {})
    n2 = PipelineNode("n2", {})
    g.add_node(n1)
    g.add_node(n2)
    edge = PipelineEdge(n1.name, n2.name, {})
    g.add_edge(edge)
    assert g.is_cyclic() is False

    edge = PipelineEdge(n2.name, n1.name, {})
    g.add_edge(edge)
    assert g.is_cyclic() is True


def test_graph_set_node(graph: PipelineGraph[PipelineNode, PipelineEdge]) -> None:
    new_node = PipelineNode("n1", {})
    graph.set_node(new_node)
    new_node_from_graph = graph.get_node_by_name("n1")
    assert new_node_from_graph.parents == []
    assert new_node_from_graph.children == ["n2"]


def test_graph_validate_edge_bad_node_name(
    graph: PipelineGraph[PipelineNode, PipelineEdge],
) -> None:
    with pytest.raises(KeyError):
        graph.add_edge(PipelineEdge("n0", "n1", {}))
    with pytest.raises(KeyError):
        graph.add_edge(PipelineEdge("n1", "n12", {}))


def test_graph_validate_edge_no_parallel_edges(
    graph: PipelineGraph[PipelineNode, PipelineEdge],
) -> None:
    with pytest.raises(ValueError):
        graph.add_edge(PipelineEdge("n1", "n2", {}))



from neo4j_graphrag.llm import CohereLLM, LLMResponse

# set api key here on in the CO_API_KEY env var
api_key = None

llm = CohereLLM(
    model_name="command-r",
    api_key=api_key,
)
res: LLMResponse = llm.invoke("say something")
print(res.content)



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import importlib
import sys
from abc import ABC
from typing import ClassVar
from unittest.mock import patch

import neo4j
import pytest
from neo4j_graphrag.embeddings import Embedder, OpenAIEmbeddings
from neo4j_graphrag.experimental.pipeline import Pipeline
from neo4j_graphrag.experimental.pipeline.config.object_config import (
    EmbedderConfig,
    EmbedderType,
    LLMConfig,
    LLMType,
    Neo4jDriverConfig,
    Neo4jDriverType,
    ObjectConfig,
)
from neo4j_graphrag.llm import LLMInterface, OpenAILLM


def test_get_class_no_optional_module() -> None:
    c: ObjectConfig[object] = ObjectConfig()
    klass = c._get_class("neo4j_graphrag.experimental.pipeline.Pipeline")
    assert klass == Pipeline


def test_get_class_optional_module() -> None:
    c: ObjectConfig[object] = ObjectConfig()
    klass = c._get_class(
        "Pipeline", optional_module="neo4j_graphrag.experimental.pipeline"
    )
    assert klass == Pipeline


def test_get_class_path_and_optional_module() -> None:
    c: ObjectConfig[object] = ObjectConfig()
    klass = c._get_class(
        "pipeline.Pipeline", optional_module="neo4j_graphrag.experimental"
    )
    assert klass == Pipeline


def test_get_class_wrong_path() -> None:
    c: ObjectConfig[object] = ObjectConfig()
    with pytest.raises(ValueError):
        c._get_class("MyClass")


class _MyClass:
    def __init__(self, param: str) -> None:
        self.param = param


class _MyInterface(ABC): ...


def test_parse_after_module_reload() -> None:
    class MyClassConfig(ObjectConfig[_MyClass]):
        DEFAULT_MODULE: ClassVar[str] = __name__
        INTERFACE: ClassVar[type] = _MyClass

    param_value = "value"
    config = MyClassConfig.model_validate(
        {"class_": f"{__name__}.{_MyClass.__name__}", "params_": {"param": param_value}}
    )
    importlib.reload(sys.modules[__name__])

    my_obj = config.parse()
    assert isinstance(my_obj, _MyClass)
    assert my_obj.param == param_value


def test_neo4j_driver_config() -> None:
    config = Neo4jDriverConfig.model_validate(
        {
            "params_": {
                "uri": "bolt://",
                "user": "a user",
                "password": "a password",
            }
        }
    )
    assert config.class_ == "not used"
    assert config.params_ == {
        "uri": "bolt://",
        "user": "a user",
        "password": "a password",
    }
    with patch(
        "neo4j_graphrag.experimental.pipeline.config.object_config.neo4j.GraphDatabase.driver"
    ) as driver_mock:
        driver_mock.return_value = "a driver"
        d = config.parse()
        driver_mock.assert_called_once_with("bolt://", auth=("a user", "a password"))
        assert d == "a driver"  # type: ignore


def test_neo4j_driver_type_with_driver(driver: neo4j.Driver) -> None:
    driver_type = Neo4jDriverType(driver)
    assert driver_type.parse() == driver


def test_neo4j_driver_type_with_config() -> None:
    driver_type = Neo4jDriverType(
        Neo4jDriverConfig(
            params_={
                "uri": "bolt://",
                "user": "",
                "password": "",
            }
        )
    )
    driver = driver_type.parse()
    assert isinstance(driver, neo4j.Driver)


def test_llm_config() -> None:
    config = LLMConfig.model_validate(
        {
            "class_": "OpenAILLM",
            "params_": {"model_name": "gpt-4o", "api_key": "my-api-key"},
        }
    )
    assert config.class_ == "OpenAILLM"
    assert config.get_module() == "neo4j_graphrag.llm"
    assert config.get_interface() == LLMInterface
    assert config.params_ == {"model_name": "gpt-4o", "api_key": "my-api-key"}
    d = config.parse()
    assert isinstance(d, OpenAILLM)


def test_llm_type_with_driver(llm: LLMInterface) -> None:
    llm_type = LLMType(llm)
    assert llm_type.parse() == llm


def test_llm_type_with_config() -> None:
    llm_type = LLMType(
        LLMConfig(
            class_="OpenAILLM",
            params_={"model_name": "gpt-4o", "api_key": "my-api-key"},
        )
    )
    llm = llm_type.parse()
    assert isinstance(llm, OpenAILLM)


def test_embedder_config() -> None:
    config = EmbedderConfig.model_validate(
        {
            "class_": "OpenAIEmbeddings",
            "params_": {"api_key": "my-api-key"},
        }
    )
    assert config.class_ == "OpenAIEmbeddings"
    assert config.get_module() == "neo4j_graphrag.embeddings"
    assert config.get_interface() == Embedder
    assert config.params_ == {"api_key": "my-api-key"}
    d = config.parse()
    assert isinstance(d, OpenAIEmbeddings)


def test_embedder_type_with_embedder(embedder: Embedder) -> None:
    embedder_type = EmbedderType(embedder)
    assert embedder_type.parse() == embedder


def test_embedder_type_with_config() -> None:
    embedder_type = EmbedderType(
        EmbedderConfig(
            class_="OpenAIEmbeddings",
            params_={"api_key": "my-api-key"},
        )
    )
    embedder = embedder_type.parse()
    assert isinstance(embedder, OpenAIEmbeddings)



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from __future__ import annotations

import inspect
import types
from abc import ABC, ABCMeta, abstractmethod
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Optional,
    TypeVar,
    get_args,
    get_origin,
    Union,
    Dict,
    get_type_hints,
)

import neo4j
from typing_extensions import ParamSpec

from neo4j_graphrag.exceptions import Neo4jVersionError
from neo4j_graphrag.types import RawSearchResult, RetrieverResult, RetrieverResultItem
from neo4j_graphrag.utils.version_utils import (
    get_version,
    has_metadata_filtering_support,
    has_vector_index_support,
    is_version_5_23_or_above,
)
from neo4j_graphrag.utils import driver_config

if TYPE_CHECKING:
    from neo4j_graphrag.tool import (
        ObjectParameter,
        Tool,
        ToolParameter,
    )

T = ParamSpec("T")
P = TypeVar("P")


def copy_function(f: Callable[T, P]) -> Callable[T, P]:
    """Based on https://stackoverflow.com/a/30714299"""
    g = types.FunctionType(
        f.__code__,
        f.__globals__,
        name=f.__name__,
        argdefs=f.__defaults__,
        closure=f.__closure__,
    )
    # in case f was given attrs (note this dict is a shallow copy):
    g.__dict__.update(f.__dict__)
    return g


class RetrieverMetaclass(ABCMeta):
    """This metaclass is used to copy the docstring from the
    `get_search_results` method, instantiated in all subclasses,
    to the `search` method in the base class.
    """

    def __new__(
        meta, name: str, bases: tuple[type, ...], attrs: dict[str, Any]
    ) -> type:
        if "search" in attrs:
            # search method was explicitly overridden, do nothing
            return type.__new__(meta, name, bases, attrs)
        # otherwise, we copy the signature and doc of the get_search_results
        # method to a copy of the search method
        get_search_results_method = attrs.get("get_search_results")
        search_method = None
        for b in bases:
            search_method = getattr(b, "search", None)
            if search_method is not None:
                break
        if search_method and get_search_results_method:
            new_search_method = copy_function(search_method)
            new_search_method.__doc__ = get_search_results_method.__doc__
            new_search_method.__signature__ = inspect.signature(  # type: ignore
                get_search_results_method
            )
            attrs["search"] = new_search_method
        return type.__new__(meta, name, bases, attrs)


class Retriever(ABC, metaclass=RetrieverMetaclass):
    """
    Abstract class for Neo4j retrievers
    """

    index_name: str
    VERIFY_NEO4J_VERSION = True

    def __init__(self, driver: neo4j.Driver, neo4j_database: Optional[str] = None):
        self.driver = driver_config.override_user_agent(driver)
        self.neo4j_database = neo4j_database
        if self.VERIFY_NEO4J_VERSION:
            version_tuple, is_aura, _ = get_version(self.driver, self.neo4j_database)
            self.neo4j_version_is_5_23_or_above = is_version_5_23_or_above(
                version_tuple
            )
            if not has_vector_index_support(
                version_tuple
            ) or not has_metadata_filtering_support(version_tuple, is_aura):
                raise Neo4jVersionError()

    def _fetch_index_infos(self, vector_index_name: str) -> None:
        """Fetch the node label and embedding property from the index definition

        Args:
            vector_index_name (str): Name of the vector index
        """
        query = (
            "SHOW VECTOR INDEXES "
            "YIELD name, labelsOrTypes, properties, options "
            "WHERE name = $index_name "
            "RETURN labelsOrTypes as labels, properties, "
            "options.indexConfig.`vector.dimensions` as dimensions"
        )
        query_result = self.driver.execute_query(
            query,
            {"index_name": vector_index_name},
            database_=self.neo4j_database,
            routing_=neo4j.RoutingControl.READ,
        )
        try:
            result = query_result.records[0]
            self._node_label = result["labels"][0]
            self._embedding_node_property = result["properties"][0]
            self._embedding_dimension = result["dimensions"]
        except IndexError as e:
            raise Exception(f"No index with name {self.index_name} found") from e

    def search(self, *args: Any, **kwargs: Any) -> RetrieverResult:
        """Search method. Call the `get_search_results` method that returns
        a list of `neo4j.Record`, and format them using the function returned by
        `get_result_formatter` to return `RetrieverResult`.
        """
        raw_result = self.get_search_results(*args, **kwargs)
        formatter = self.get_result_formatter()
        search_items = [formatter(record) for record in raw_result.records]
        metadata = raw_result.metadata or {}
        metadata["__retriever"] = self.__class__.__name__
        return RetrieverResult(
            items=search_items,
            metadata=metadata,
        )

    @abstractmethod
    def get_search_results(self, *args: Any, **kwargs: Any) -> RawSearchResult:
        """This method must be implemented in each child class. It will
        receive the same parameters provided to the public interface via
        the `search` method, after validation. It returns a `RawSearchResult`
        object which comprises a list of `neo4j.Record` objects and an optional
        `metadata` dictionary that can contain retriever-level information.

        Note that, even though this method is not intended to be called from
        outside the class, we make it public to make it clearer for the developers
        that it should be implemented in child classes.

        Returns:
            RawSearchResult: List of Neo4j Records and optional metadata dict
        """
        pass

    def get_result_formatter(self) -> Callable[[neo4j.Record], RetrieverResultItem]:
        """
        Returns the function to use to transform a neo4j.Record to a RetrieverResultItem.
        """
        if hasattr(self, "result_formatter"):
            return self.result_formatter or self.default_record_formatter
        return self.default_record_formatter

    def default_record_formatter(self, record: neo4j.Record) -> RetrieverResultItem:
        """
        Best effort to guess the node-to-text method. Inherited classes
        can override this method to implement custom text formatting.
        """
        return RetrieverResultItem(content=str(record), metadata=record.get("metadata"))

    def get_parameters(
        self, parameter_descriptions: Optional[Dict[str, str]] = None
    ) -> "ObjectParameter":
        """Return the parameters that this retriever expects for tool conversion.

        This method automatically infers parameters from the get_search_results method signature.

        Args:
            parameter_descriptions (Optional[Dict[str, str]]): Custom descriptions for parameters.
                Keys should match parameter names from get_search_results method.

        Returns:
            ObjectParameter: The parameter definition for this retriever
        """
        return self._infer_parameters_from_signature(parameter_descriptions or {})

    def _infer_parameters_from_signature(
        self, parameter_descriptions: Dict[str, str]
    ) -> "ObjectParameter":
        """Infer parameters from the get_search_results method signature."""
        # Import here to avoid circular imports
        from neo4j_graphrag.tool import (
            ObjectParameter,
        )

        # Get the method signature and resolved type hints
        sig = inspect.signature(self.get_search_results)
        try:
            type_hints = get_type_hints(self.get_search_results)
        except (NameError, AttributeError):
            # If type hints can't be resolved, fall back to annotation strings
            type_hints = {}

        properties: Dict[str, "ToolParameter"] = {}
        required_properties = []

        for param_name, param in sig.parameters.items():
            # Skip 'self' parameter
            if param_name == "self":
                continue

            # Skip **kwargs
            if param.kind == inspect.Parameter.VAR_KEYWORD:
                continue

            # Determine if parameter is required (no default value)
            is_required = param.default is inspect.Parameter.empty

            # Use resolved type hint if available, otherwise fall back to annotation
            type_annotation = type_hints.get(param_name, param.annotation)

            # Get the parameter type and create appropriate tool parameter
            tool_param = self._create_tool_parameter_from_type(
                param_name, type_annotation, is_required, parameter_descriptions
            )

            if tool_param:
                properties[param_name] = tool_param
                if is_required:
                    required_properties.append(param_name)

        return ObjectParameter(
            description=f"Parameters for {self.__class__.__name__}",
            properties=properties,
            required_properties=required_properties,
            additional_properties=False,
        )

    def _create_tool_parameter_from_type(
        self,
        param_name: str,
        type_annotation: Any,
        is_required: bool,
        parameter_descriptions: Dict[str, str],
    ) -> Optional["ToolParameter"]:
        """Create a tool parameter from a type annotation."""
        # Import here to avoid circular imports
        from neo4j_graphrag.tool import (
            StringParameter,
            IntegerParameter,
            NumberParameter,
            ArrayParameter,
            ObjectParameter,
        )

        # Handle None/missing annotation
        if type_annotation is inspect.Parameter.empty or type_annotation is None:
            return StringParameter(
                description=parameter_descriptions.get(
                    param_name, f"Parameter {param_name}"
                ),
                required=is_required,
            )

        # Get the origin and args for generic types
        origin = get_origin(type_annotation)
        args = get_args(type_annotation)

        # Handle Optional[T] and Union[T, None]
        if origin is Union:
            # Remove None from union args to get the actual type
            non_none_args = [arg for arg in args if arg is not type(None)]
            if len(non_none_args) == 1:
                # This is Optional[T], use T
                type_annotation = non_none_args[0]
                # Re-calculate origin and args for the unwrapped type
                origin = get_origin(type_annotation)
                args = get_args(type_annotation)
            elif len(non_none_args) > 1:
                # This is Union[T, U, ...], treat as string for now
                return StringParameter(
                    description=parameter_descriptions.get(
                        param_name, f"Parameter {param_name}"
                    ),
                    required=is_required,
                )

        # Handle specific types
        if type_annotation is str:
            return StringParameter(
                description=parameter_descriptions.get(
                    param_name, f"Parameter {param_name}"
                ),
                required=is_required,
            )
        elif type_annotation is int:
            return IntegerParameter(
                description=parameter_descriptions.get(
                    param_name, f"Parameter {param_name}"
                ),
                minimum=1
                if param_name in ["top_k", "effective_search_ratio"]
                else None,
                required=is_required,
            )
        elif type_annotation is float:
            constraints: Dict[str, Any] = {}
            if param_name == "alpha":
                constraints.update(minimum=0.0, maximum=1.0)
            return NumberParameter(
                description=parameter_descriptions.get(
                    param_name, f"Parameter {param_name}"
                ),
                required=is_required,
                **constraints,
            )
        elif (
            origin is list
            or type_annotation is list
            or (
                hasattr(type_annotation, "__origin__")
                and type_annotation.__origin__ is list
            )
            or str(type_annotation).startswith("list[")
        ):
            # Handle list[float] for vectors
            if args and args[0] is float:
                return ArrayParameter(
                    items=NumberParameter(
                        description="A single vector component",
                        required=False,
                    ),
                    description=parameter_descriptions.get(
                        param_name, f"Parameter {param_name}"
                    ),
                    required=is_required,
                )
            else:
                # For complex list types like List[LLMMessage], treat as object
                return ObjectParameter(
                    description=parameter_descriptions.get(
                        param_name, f"Parameter {param_name}"
                    ),
                    properties={},
                    additional_properties=True,
                    required=is_required,
                )
        elif origin is dict or (
            hasattr(type_annotation, "__origin__")
            and type_annotation.__origin__ is dict
        ):
            return ObjectParameter(
                description=parameter_descriptions.get(
                    param_name, f"Parameter {param_name}"
                ),
                properties={},
                additional_properties=True,
                required=is_required,
            )
        else:
            # Check if it's a complex type that should be an object
            type_name = str(type_annotation)
            if any(
                keyword in type_name.lower()
                for keyword in ["dict", "list", "optional[dict", "optional[list"]
            ):
                return ObjectParameter(
                    description=parameter_descriptions.get(
                        param_name, f"Parameter {param_name}"
                    ),
                    properties={},
                    additional_properties=True,
                    required=is_required,
                )
            # For other complex types or enums, default to string
            return StringParameter(
                description=parameter_descriptions.get(
                    param_name, f"Parameter {param_name}"
                ),
                required=is_required,
            )

    def convert_to_tool(
        self,
        name: str,
        description: str,
        parameter_descriptions: Optional[Dict[str, str]] = None,
    ) -> "Tool":
        """Convert this retriever to a Tool object.

        Args:
            name (str): Name for the tool.
            description (str): Description of what the tool does.
            parameter_descriptions (Optional[Dict[str, str]]): Optional descriptions for each parameter.
                Keys should match parameter names from get_search_results method.

        Returns:
            Tool: A Tool object configured to use this retriever's search functionality.
        """
        # Import here to avoid circular imports
        from neo4j_graphrag.tool import Tool

        # Get parameters from the retriever with custom descriptions
        parameters = self.get_parameters(parameter_descriptions or {})

        # Define a function that matches the Callable[[str, ...], Any] signature
        def execute_func(**kwargs: Any) -> Any:
            return self.search(**kwargs)

        # Create a Tool object from the retriever
        return Tool(
            name=name,
            description=description,
            execute_func=execute_func,
            parameters=parameters,
        )


class ExternalRetriever(Retriever, ABC):
    """
    Abstract class for External Vector Stores
    """

    VERIFY_NEO4J_VERSION = False

    def __init__(
        self,
        driver: neo4j.Driver,
        id_property_external: str,
        id_property_neo4j: str,
        neo4j_database: Optional[str] = None,
        node_label_neo4j: Optional[str] = None,
    ):
        super().__init__(driver)
        self.id_property_external = id_property_external
        self.id_property_neo4j = id_property_neo4j
        self.node_label_neo4j = node_label_neo4j
        self.neo4j_database = neo4j_database

    @abstractmethod
    def get_search_results(
        self,
        query_vector: Optional[list[float]] = None,
        query_text: Optional[str] = None,
        top_k: int = 5,
        **kwargs: Any,
    ) -> RawSearchResult:
        """

        Returns:
                RawSearchResult: List of Neo4j Records and optional metadata dict

        """
        pass



#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Iterable,
    List,
    Optional,
    Type,
    Union,
    cast,
    overload,
)

from pydantic import BaseModel, ValidationError

from neo4j_graphrag.exceptions import LLMGenerationError
from neo4j_graphrag.llm.base import LLMInterface, LLMInterfaceV2
from neo4j_graphrag.llm.types import (
    BaseMessage,
    LLMResponse,
    MessageList,
    UserMessage,
)
from neo4j_graphrag.message_history import MessageHistory
from neo4j_graphrag.types import LLMMessage
from neo4j_graphrag.utils.rate_limit import (
    RateLimitHandler,
)
from neo4j_graphrag.utils.rate_limit import (
    async_rate_limit_handler as async_rate_limit_handler_decorator,
)
from neo4j_graphrag.utils.rate_limit import (
    rate_limit_handler as rate_limit_handler_decorator,
)

if TYPE_CHECKING:
    from anthropic import NotGiven
    from anthropic.types.message_param import MessageParam


# pylint: disable=redefined-builtin, arguments-differ, raise-missing-from, no-else-return, import-outside-toplevel
class AnthropicLLM(LLMInterface, LLMInterfaceV2):
    """Interface for large language models on Anthropic

    Args:
        model_name (str, optional): Name of the LLM to use. Defaults to "gemini-1.5-flash-001".
        model_params (Optional[dict], optional): Additional parameters for LLMInterface(V1) passed to the model when text is sent to it. Defaults to None.
        system_instruction: Optional[str], optional): Additional instructions for setting the behavior and context for the model in a conversation. Defaults to None.
        rate_limit_handler (Optional[RateLimitHandler], optional): Handler for managing rate limits for LLMInterface(V1). Defaults to None.
        **kwargs (Any): Arguments passed to the model when for the class is initialised. Defaults to None.

    Raises:
        LLMGenerationError: If there's an error generating the response from the model.

    Example:

    .. code-block:: python

        from neo4j_graphrag.llm import AnthropicLLM

        llm = AnthropicLLM(
            model_name="claude-3-opus-20240229",
            model_params={"max_tokens": 1000},
            api_key="sk...",   # can also be read from env vars
        )
        llm.invoke("Who is the mother of Paul Atreides?")
    """

    def __init__(
        self,
        model_name: str,
        model_params: Optional[dict[str, Any]] = None,
        rate_limit_handler: Optional[RateLimitHandler] = None,
        **kwargs: Any,
    ):
        try:
            import anthropic
        except ImportError:
            raise ImportError(
                """Could not import Anthropic Python client.
                Please install it with `pip install "neo4j-graphrag[anthropic]"`."""
            )
        LLMInterfaceV2.__init__(
            self,
            model_name=model_name,
            model_params=model_params or {},
            rate_limit_handler=rate_limit_handler,
            **kwargs,
        )
        self.anthropic = anthropic
        self.client = anthropic.Anthropic(**kwargs)
        self.async_client = anthropic.AsyncAnthropic(**kwargs)

    # overloads for LLMInterface and LLMInterfaceV2 methods
    @overload  # type: ignore[no-overload-impl]
    def invoke(
        self,
        input: str,
        message_history: Optional[Union[List[LLMMessage], MessageHistory]] = None,
        system_instruction: Optional[str] = None,
    ) -> LLMResponse: ...

    @overload
    def invoke(
        self,
        input: List[LLMMessage],
        response_format: Optional[Union[Type[BaseModel], dict[str, Any]]] = None,
        **kwargs: Any,
    ) -> LLMResponse: ...

    @overload  # type: ignore[no-overload-impl]
    async def ainvoke(
        self,
        input: str,
        message_history: Optional[Union[List[LLMMessage], MessageHistory]] = None,
        system_instruction: Optional[str] = None,
    ) -> LLMResponse: ...

    @overload
    async def ainvoke(
        self,
        input: List[LLMMessage],
        response_format: Optional[Union[Type[BaseModel], dict[str, Any]]] = None,
        **kwargs: Any,
    ) -> LLMResponse: ...

    # switching logics to LLMInterface or LLMInterfaceV2
    def invoke(  # type: ignore[no-redef]
        self,
        input: Union[str, List[LLMMessage]],
        message_history: Optional[Union[List[LLMMessage], MessageHistory]] = None,
        system_instruction: Optional[str] = None,
        response_format: Optional[Union[Type[BaseModel], dict[str, Any]]] = None,
        **kwargs: Any,
    ) -> LLMResponse:
        if isinstance(input, str):
            return self.__invoke_v1(input, message_history, system_instruction)
        elif isinstance(input, list):
            return self.__invoke_v2(input, response_format=response_format, **kwargs)
        else:
            raise ValueError(f"Invalid input type for invoke method - {type(input)}")

    async def ainvoke(  # type: ignore[no-redef]
        self,
        input: Union[str, List[LLMMessage]],
        message_history: Optional[Union[List[LLMMessage], MessageHistory]] = None,
        system_instruction: Optional[str] = None,
        response_format: Optional[Union[Type[BaseModel], dict[str, Any]]] = None,
        **kwargs: Any,
    ) -> LLMResponse:
        if isinstance(input, str):
            return await self.__ainvoke_v1(input, message_history, system_instruction)
        elif isinstance(input, list):
            return await self.__ainvoke_v2(
                input, response_format=response_format, **kwargs
            )
        else:
            raise ValueError(f"Invalid input type for ainvoke method - {type(input)}")

    # implementaions
    @rate_limit_handler_decorator
    def __invoke_v1(
        self,
        input: str,
        message_history: Optional[Union[List[LLMMessage], MessageHistory]] = None,
        system_instruction: Optional[str] = None,
    ) -> LLMResponse:
        """Sends text to the LLM and returns a response.

        Args:
            input (str): The text to send to the LLM.
            message_history (Optional[Union[List[LLMMessage], MessageHistory]]): A collection previous messages,
                with each message having a specific role assigned.
            system_instruction (Optional[str]): An option to override the llm system message for this invocation.

        Returns:
            LLMResponse: The response from the LLM.
        """
        try:
            if isinstance(message_history, MessageHistory):
                message_history = message_history.messages
            messages = self.get_messages(input, message_history)
            response = self.client.messages.create(
                model=self.model_name,
                system=system_instruction or self.anthropic.NOT_GIVEN,
                messages=messages,
                **self.model_params,
            )
            response_content = response.content
            if response_content and len(response_content) > 0:
                text = response_content[0].text
            else:
                raise LLMGenerationError("LLM returned empty response.")
            return LLMResponse(content=text)
        except self.anthropic.APIError as e:
            raise LLMGenerationError(e)

    @rate_limit_handler_decorator
    def __invoke_v2(
        self,
        input: List[LLMMessage],
        response_format: Optional[Union[Type[BaseModel], dict[str, Any]]] = None,
        **kwargs: Any,
    ) -> LLMResponse:
        if response_format is not None:
            raise NotImplementedError(
                "AnthropicLLM does not currently support structured output"
            )
        try:
            system_instruction, messages = self.get_messages_v2(input)
            response = self.client.messages.create(
                model=self.model_name,
                system=system_instruction,
                messages=messages,
                **self.model_params,
                **kwargs,
            )
            response_content = response.content
            if response_content and len(response_content) > 0:
                text = response_content[0].text
            else:
                raise LLMGenerationError("LLM returned empty response.")
            return LLMResponse(content=text)
        except self.anthropic.APIError as e:
            raise LLMGenerationError(e)

    @async_rate_limit_handler_decorator
    async def __ainvoke_v1(
        self,
        input: str,
        message_history: Optional[Union[List[LLMMessage], MessageHistory]] = None,
        system_instruction: Optional[str] = None,
    ) -> LLMResponse:
        """Asynchronously sends text to the LLM and returns a response.

        Args:
            input (str): The text to send to the LLM.
            message_history (Optional[Union[List[LLMMessage], MessageHistory]]): A collection previous messages,
                with each message having a specific role assigned.
            system_instruction (Optional[str]): An option to override the llm system message for this invocation.

        Returns:
            LLMResponse: The response from the LLM.
        """
        try:
            if isinstance(message_history, MessageHistory):
                message_history = message_history.messages
            messages = self.get_messages(input, message_history)
            response = await self.async_client.messages.create(
                model=self.model_name,
                system=system_instruction or self.anthropic.NOT_GIVEN,
                messages=messages,
                **self.model_params,
            )
            response_content = response.content
            if response_content and len(response_content) > 0:
                text = response_content[0].text
            else:
                raise LLMGenerationError("LLM returned empty response.")
            return LLMResponse(content=text)
        except self.anthropic.APIError as e:
            raise LLMGenerationError(e)

    @async_rate_limit_handler_decorator
    async def __ainvoke_v2(
        self,
        input: List[LLMMessage],
        response_format: Optional[Union[Type[BaseModel], dict[str, Any]]] = None,
        **kwargs: Any,
    ) -> LLMResponse:
        """Asynchronously sends text to the LLM and returns a response.

        Args:
            input (List[LLMMessage]): The messages to send to the LLM.
            response_format: Not supported by AnthropicLLM.

        Returns:
            LLMResponse: The response from the LLM.
        """
        if response_format is not None:
            raise NotImplementedError(
                "AnthropicLLM does not currently support structured output"
            )
        try:
            system_instruction, messages = self.get_messages_v2(input)
            response = await self.async_client.messages.create(
                model=self.model_name,
                system=system_instruction,
                messages=messages,
                **self.model_params,
                **kwargs,
            )
            response_content = response.content
            if response_content and len(response_content) > 0:
                text = response_content[0].text
            else:
                raise LLMGenerationError("LLM returned empty response.")
            return LLMResponse(content=text)
        except self.anthropic.APIError as e:
            raise LLMGenerationError(e)

    # subsidiary methods
    def get_messages(
        self,
        input: str,
        message_history: Optional[Union[List[LLMMessage], MessageHistory]] = None,
    ) -> Iterable[MessageParam]:
        """Constructs the message list for the LLM from the input and message history."""
        messages: list[dict[str, str]] = []
        if message_history:
            if isinstance(message_history, MessageHistory):
                message_history = message_history.messages
            try:
                MessageList(messages=cast(list[BaseMessage], message_history))
            except ValidationError as e:
                raise LLMGenerationError(e.errors()) from e
            messages.extend(cast(Iterable[dict[str, Any]], message_history))
        messages.append(UserMessage(content=input).model_dump())
        return messages  # type: ignore

    def get_messages_v2(
        self,
        input: list[LLMMessage],
    ) -> tuple[Union[str, NotGiven], Iterable[MessageParam]]:
        """Constructs the message list for the LLM from the input."""
        messages: list[MessageParam] = []
        system_instruction: Union[str, NotGiven] = self.anthropic.NOT_GIVEN
        for i in input:
            if i["role"] == "system":
                system_instruction = i["content"]
            else:
                if i["role"] not in ("user", "assistant"):
                    raise ValueError(f"Unknown role: {i['role']}")
                messages.append(
                    self.anthropic.types.MessageParam(
                        role=i["role"],
                        content=i["content"],
                    )
                )
        return system_instruction, messages



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import Any
from unittest.mock import patch

import pytest
from pydantic import ValidationError

from neo4j_viz import VisualizationGraph
from neo4j_graphrag.experimental.components.schema import GraphSchema
from neo4j_graphrag.experimental.utils.schema import schema_visualization


@pytest.fixture(scope="module")
def valid_schema_dict() -> dict[str, Any]:
    return {
        "node_types": [
            "Location",
            {
                "label": "Person",
                "properties": [
                    {"name": "name", "type": "STRING", "required": True},
                    {"name": "birthYear", "type": "INTEGER"},
                ],
            },
        ],
        "relationship_types": [
            "BORN_IN",
            {
                "label": "KNOWS",
                "properties": [
                    {"name": "since", "type": "LOCAL_DATETIME"},
                ],
            },
        ],
        "patterns": [
            ("Person", "BORN_IN", "Location"),
            ("Person", "KNOWS", "Person"),
        ],
    }


@pytest.fixture(scope="module")
def invalid_schema_dict() -> dict[str, Any]:
    return {
        "node_types": [
            {
                "label": "Person",
                "properties": [
                    {"name": "name", "type": "STRING", "required": True},
                    {"name": "birthYear", "type": "INTEGER"},
                ],
            },
        ],
        "relationship_types": [
            "BORN_IN",
        ],
        "patterns": [
            (
                "Person",
                "BORN_IN",
                "Location",
            ),  # invalid pattern, "Location" node type not defined
        ],
    }


@patch("neo4j_graphrag.experimental.utils.schema.VisualizationGraph", None)
def test_schema_visualization_import_error() -> None:
    with pytest.raises(ImportError):
        schema_visualization({})


def test_schema_visualization_invalid_schema_dict(
    invalid_schema_dict: dict[str, Any],
) -> None:
    with pytest.raises(ValidationError):
        schema_visualization(invalid_schema_dict)


def test_schema_visualization_valid_schema_dict(
    valid_schema_dict: dict[str, Any],
) -> None:
    g = schema_visualization(valid_schema_dict)
    assert isinstance(g, VisualizationGraph)
    assert len(g.nodes) == 2
    assert len(g.relationships) == 2


def test_schema_visualization_schema_object(valid_schema_dict: dict[str, Any]) -> None:
    schema = GraphSchema.model_validate(valid_schema_dict)
    g = schema_visualization(schema)
    assert isinstance(g, VisualizationGraph)
    assert len(g.nodes) == 2
    assert len(g.relationships) == 2



"""Implement a custom writer to save the results, for instance by using
custom Cypher queries.
"""

import neo4j
from neo4j_graphrag.experimental.components.kg_writer import KGWriter, KGWriterModel
from neo4j_graphrag.experimental.components.types import LexicalGraphConfig, Neo4jGraph
from pydantic import validate_call
from neo4j_graphrag.utils import driver_config


class MyWriter(KGWriter):
    def __init__(self, driver: neo4j.Driver) -> None:
        self.driver = driver_config.override_user_agent(driver)

    @validate_call
    async def run(
        self,
        graph: Neo4jGraph,
        lexical_graph_config: LexicalGraphConfig = LexicalGraphConfig(),
    ) -> KGWriterModel:
        try:
            self.driver.execute_query("my query")
            return KGWriterModel(status="SUCCESS")
        except Exception:
            return KGWriterModel(status="FAILURE")



import warnings
from typing import Any, Dict, List, Literal, Optional

from pydantic import BaseModel

from neo4j_graphrag.types import LLMMessage as _LLMMessage


def __getattr__(name: str) -> Any:
    if name == "LLMMessage":
        warnings.warn(
            "LLMMessage has been moved to neo4j_graphrag.types. Please update your imports.",
            DeprecationWarning,
            stacklevel=2,
        )
        return _LLMMessage
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


class LLMResponse(BaseModel):
    content: str


class BaseMessage(BaseModel):
    role: Literal["user", "assistant", "system"]
    content: str


class UserMessage(BaseMessage):
    role: Literal["user"] = "user"


class SystemMessage(BaseMessage):
    role: Literal["system"] = "system"


class MessageList(BaseModel):
    messages: list[BaseMessage]


class ToolCall(BaseModel):
    """A tool call made by an LLM."""

    name: str
    arguments: Dict[str, Any]


class ToolCallResponse(BaseModel):
    """Response from an LLM containing tool calls."""

    tool_calls: List[ToolCall]
    content: Optional[str] = None



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from __future__ import annotations

import logging
import uuid
from datetime import date, datetime, time
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, ConfigDict, Field

from neo4j_graphrag.experimental.pipeline.component import DataModel


logger = logging.getLogger(__name__)


class GeoPoint(BaseModel):
    """Represents a geographic point with latitude, longitude, and height.

    Attributes:
        latitude (float): The latitude coordinate.
        longitude (float): The longitude coordinate.
        height (float): The height/altitude coordinate.
    """

    model_config = ConfigDict(extra="forbid")

    latitude: float
    longitude: float
    height: float


# Define primitive value types
PrimitiveValue = Union[bool, int, float, str]

# Define temporal value types (ISO 8601 strings for date, time, datetime)
TemporalValue = Union[date, time, datetime]

# Define duration as a string with ISO 8601 duration pattern (e.g., "P1Y2M3DT4H5M6S")
Duration = str

# Define the complete PropertyValue union covering all Neo4j property types
PropertyValue = Union[
    PrimitiveValue,
    TemporalValue,
    Duration,
    List[Union[bool, int, float, str]],  # Arrays of primitives
    GeoPoint,
]


class DocumentInfo(DataModel):
    """A document loaded by a DataLoader.

    Attributes:
        path (str): Document path.
        metadata (Optional[dict[str, Any]]): Metadata associated with this document.
        uid (str): Unique identifier for this document.
    """

    path: str
    metadata: Optional[Dict[str, str]] = None
    uid: str = Field(default_factory=lambda: str(uuid.uuid4()))
    document_type: Optional[str] = None

    @property
    def document_id(self) -> str:
        return self.uid


class PdfDocument(DataModel):
    text: str
    document_info: DocumentInfo


class TextChunk(BaseModel):
    """A chunk of text split from a document by a text splitter.

    Attributes:
        text (str): The raw chunk text.
        index (int): The position of this chunk in the original document.
        metadata (Optional[dict[str, Any]]): Metadata associated with this chunk.
        uid (str): Unique identifier for this chunk.
    """

    text: str
    index: int
    metadata: Optional[dict[str, Any]] = None
    uid: str = Field(default_factory=lambda: str(uuid.uuid4()))

    @property
    def chunk_id(self) -> str:
        return self.uid


class TextChunks(DataModel):
    """A collection of text chunks returned from a text splitter.

    Attributes:
        chunks (list[TextChunk]): A list of text chunks.
    """

    chunks: list[TextChunk]


class Neo4jNode(BaseModel):
    """Represents a Neo4j node.

    Attributes:
        id (str): The ID of the node. This ID is used to refer to the node for relationship creation.
        label (str): The label of the node.
        properties (dict[str, PropertyValue]): A dictionary of properties attached to the node.
        embedding_properties (dict[str, list[float]]): A dictionary of embedding properties attached to the node.
    """

    model_config = ConfigDict(extra="forbid")

    id: str
    label: str
    properties: dict[str, PropertyValue] = Field(default_factory=dict)
    embedding_properties: dict[str, list[float]] = Field(default_factory=dict)

    @property
    def token(self) -> str:
        return self.label


class Neo4jRelationship(BaseModel):
    """Represents a Neo4j relationship.

    Attributes:
        start_node_id (str): The ID of the start node.
        end_node_id (str): The ID of the end node.
        type (str): The relationship type.
        properties (dict[str, PropertyValue]): A dictionary of properties attached to the relationship.
        embedding_properties (dict[str, list[float]]): A dictionary of embedding properties attached to the relationship.
    """

    model_config = ConfigDict(extra="forbid")

    start_node_id: str
    end_node_id: str
    type: str
    properties: dict[str, PropertyValue] = Field(default_factory=dict)
    embedding_properties: dict[str, list[float]] = Field(default_factory=dict)

    @property
    def token(self) -> str:
        return self.type


class Neo4jGraph(DataModel):
    """Represents a Neo4j graph.

    Attributes:
        nodes (list[Neo4jNode]): A list of nodes in the graph.
        relationships (list[Neo4jRelationship]): A list of relationships in the graph.
    """

    model_config = ConfigDict(extra="forbid")

    nodes: list[Neo4jNode] = Field(default_factory=list)
    relationships: list[Neo4jRelationship] = Field(default_factory=list)

    @classmethod
    def model_json_schema(cls, **kwargs: Any) -> dict[str, Any]:  # type: ignore[override]
        """Override to mark nodes and relationships as required in JSON schema."""
        schema = super().model_json_schema(**kwargs)
        # Force nodes and relationships to be required for structured output compatibility
        schema["required"] = ["nodes", "relationships"]
        return schema


class ResolutionStats(DataModel):
    number_of_nodes_to_resolve: int
    number_of_created_nodes: Optional[int] = None


DEFAULT_DOCUMENT_NODE_LABEL = "Document"
DEFAULT_CHUNK_NODE_LABEL = "Chunk"
DEFAULT_CHUNK_TO_DOCUMENT_RELATIONSHIP_TYPE = "FROM_DOCUMENT"
DEFAULT_NEXT_CHUNK_RELATIONSHIP_TYPE = "NEXT_CHUNK"
DEFAULT_NODE_TO_CHUNK_RELATIONSHIP_TYPE = "FROM_CHUNK"
DEFAULT_CHUNK_ID_PROPERTY = "id"
DEFAULT_CHUNK_INDEX_PROPERTY = "index"
DEFAULT_CHUNK_TEXT_PROPERTY = "text"
DEFAULT_CHUNK_EMBEDDING_PROPERTY = "embedding"


class LexicalGraphConfig(BaseModel):
    """Configure all labels and property names in the lexical graph."""

    id_prefix: str = Field(deprecated=True, default="")
    document_node_label: str = DEFAULT_DOCUMENT_NODE_LABEL
    chunk_node_label: str = DEFAULT_CHUNK_NODE_LABEL
    chunk_to_document_relationship_type: str = (
        DEFAULT_CHUNK_TO_DOCUMENT_RELATIONSHIP_TYPE
    )
    next_chunk_relationship_type: str = DEFAULT_NEXT_CHUNK_RELATIONSHIP_TYPE
    node_to_chunk_relationship_type: str = DEFAULT_NODE_TO_CHUNK_RELATIONSHIP_TYPE
    chunk_id_property: str = DEFAULT_CHUNK_ID_PROPERTY
    chunk_index_property: str = DEFAULT_CHUNK_INDEX_PROPERTY
    chunk_text_property: str = DEFAULT_CHUNK_TEXT_PROPERTY
    chunk_embedding_property: str = DEFAULT_CHUNK_EMBEDDING_PROPERTY

    @property
    def lexical_graph_node_labels(self) -> tuple[str, ...]:
        return self.document_node_label, self.chunk_node_label

    @property
    def lexical_graph_relationship_types(self) -> tuple[str, ...]:
        return (
            self.chunk_to_document_relationship_type,
            self.next_chunk_relationship_type,
            self.node_to_chunk_relationship_type,
        )


class GraphResult(DataModel):
    graph: Neo4jGraph
    config: LexicalGraphConfig



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from types import SimpleNamespace
from typing import Optional
from unittest.mock import MagicMock

import neo4j
import pytest
from neo4j_graphrag.exceptions import RetrieverInitializationError
from neo4j_graphrag.retrievers import WeaviateNeo4jRetriever
from neo4j_graphrag.retrievers.external.utils import get_match_query
from neo4j_graphrag.types import RetrieverResult, RetrieverResultItem
from weaviate.client import WeaviateClient


# Weaviate class with fake methods
class WClient(WeaviateClient):
    def __init__(
        self,
        node_id_value: Optional[str] = None,
        node_match_score: Optional[float] = None,
    ) -> None:
        self.collections = MagicMock()
        self.collections.get = MagicMock()
        query = MagicMock()
        self.collections.get.return_value = SimpleNamespace(query=query)
        query.near_text.return_value = SimpleNamespace(
            objects=[
                SimpleNamespace(
                    properties={"neo4j_id": node_id_value},
                    metadata=SimpleNamespace(certainty=node_match_score),
                )
            ]
        )


def test_text_search_remote_vector_store_happy_path(driver: MagicMock) -> None:
    query_text = "may thy knife chip and shatter"
    top_k = 5
    node_id_value = "node-test-id"
    node_match_score = 0.9

    wc = WClient(node_id_value=node_id_value, node_match_score=node_match_score)

    retriever = WeaviateNeo4jRetriever(
        driver=driver,
        client=wc,
        collection="dummy-collection",
        id_property_neo4j="sync_id",
        id_property_external="neo4j_id",
    )
    driver.execute_query.return_value = [
        [neo4j.Record({"node": {"sync_id": node_id_value}, "score": node_match_score})],
        None,
        None,
    ]
    search_query = get_match_query()

    records = retriever.search(query_text=query_text, top_k=top_k)

    driver.execute_query.assert_called_once_with(
        search_query,
        {
            "match_params": [
                (node_id_value, node_match_score),
            ],
            "id_property": "sync_id",
        },
        database_=None,
        routing_=neo4j.RoutingControl.READ,
    )
    assert records == RetrieverResult(
        items=[
            RetrieverResultItem(
                content="<Record node={'sync_id': 'node-test-id'} score=0.9>",
                metadata=None,
            ),
        ],
        metadata={"__retriever": "WeaviateNeo4jRetriever"},
    )


def test_invalid_neo4j_database_name(driver: MagicMock) -> None:
    node_id_value = "node-test-id"
    node_match_score = 0.9

    wc = WClient(node_id_value=node_id_value, node_match_score=node_match_score)

    with pytest.raises(RetrieverInitializationError) as exc_info:
        WeaviateNeo4jRetriever(
            driver=driver,
            client=wc,
            collection="dummy-collection",
            id_property_neo4j="sync_id",
            id_property_external="neo4j_id",
            neo4j_database=42,  # type: ignore
        )

    assert "neo4j_database" in str(exc_info.value)
    assert "Input should be a valid string" in str(exc_info.value)


def test_text_search_remote_vector_store_return_properties(driver: MagicMock) -> None:
    query_text = "may thy knife chip and shatter"
    top_k = 5
    node_id_value = "node-test-id"
    node_match_score = 0.9

    wc = WClient(node_id_value=node_id_value, node_match_score=node_match_score)

    retriever = WeaviateNeo4jRetriever(
        driver=driver,
        client=wc,
        collection="dummy-collection",
        id_property_neo4j="sync_id",
        id_property_external="neo4j_id",
        return_properties=["sync_id"],
    )
    driver.execute_query.return_value = [
        [neo4j.Record({"node": {"sync_id": node_id_value}, "score": node_match_score})],
        None,
        None,
    ]
    search_query = get_match_query(return_properties=["sync_id"])

    records = retriever.search(query_text=query_text, top_k=top_k)

    driver.execute_query.assert_called_once_with(
        search_query,
        {
            "match_params": [
                (node_id_value, node_match_score),
            ],
            "id_property": "sync_id",
        },
        database_=None,
        routing_=neo4j.RoutingControl.READ,
    )
    assert records == RetrieverResult(
        items=[
            RetrieverResultItem(
                content="<Record node={'sync_id': 'node-test-id'} score=0.9>",
                metadata=None,
            ),
        ],
        metadata={"__retriever": "WeaviateNeo4jRetriever"},
    )


def test_text_search_remote_vector_store_retrieval_query(driver: MagicMock) -> None:
    query_text = "may thy knife chip and shatter"
    top_k = 5
    node_id_value = "node-test-id"
    node_match_score = 0.9
    retrieval_query = "WITH node MATCH (node)--(m) RETURN n, m LIMIT 10"

    wc = WClient(node_id_value=node_id_value, node_match_score=node_match_score)

    retriever = WeaviateNeo4jRetriever(
        driver=driver,
        client=wc,
        collection="dummy-collection",
        id_property_neo4j="sync_id",
        id_property_external="neo4j_id",
        retrieval_query=retrieval_query,
    )
    driver.execute_query.return_value = [
        [neo4j.Record({"node": {"sync_id": node_id_value}, "score": node_match_score})],
        None,
        None,
    ]
    search_query = get_match_query(retrieval_query=retrieval_query)

    records = retriever.search(query_text=query_text, top_k=top_k)

    driver.execute_query.assert_called_once_with(
        search_query,
        {
            "match_params": [
                (node_id_value, node_match_score),
            ],
            "id_property": "sync_id",
        },
        database_=None,
        routing_=neo4j.RoutingControl.READ,
    )

    assert records == RetrieverResult(
        items=[
            RetrieverResultItem(
                content="<Record node={'sync_id': 'node-test-id'} score=0.9>",
                metadata=None,
            ),
        ],
        metadata={"__retriever": "WeaviateNeo4jRetriever"},
    )


def test_match_query() -> None:
    match_query = get_match_query()
    expected = (
        "UNWIND $match_params AS match_param "
        "WITH match_param[0] AS match_id_value, match_param[1] AS score "
        "MATCH (node) "
        "WHERE node[$id_property] = match_id_value "
        "RETURN node, score"
    )
    assert match_query.strip() == expected.strip()


def test_match_query_with_return_properties() -> None:
    match_query = get_match_query(return_properties=["name", "age"])
    expected = (
        "UNWIND $match_params AS match_param "
        "WITH match_param[0] AS match_id_value, match_param[1] AS score "
        "MATCH (node) "
        "WHERE node[$id_property] = match_id_value "
        "RETURN node {.name, .age} AS node, labels(node) AS nodeLabels, elementId(node) AS elementId, elementId(node) AS id, score"
    )
    assert match_query.strip() == expected.strip()


def test_match_query_with_retrieval_query() -> None:
    retrieval_query = "WITH node MATCH (node)--(m) RETURN node, m LIMIT 10"
    match_query = get_match_query(retrieval_query=retrieval_query)
    expected = (
        "UNWIND $match_params AS match_param "
        "WITH match_param[0] AS match_id_value, match_param[1] AS score "
        "MATCH (node) "
        "WHERE node[$id_property] = match_id_value " + retrieval_query
    )
    assert match_query.strip() == expected.strip()


def test_match_query_with_both_return_properties_and_retrieval_query() -> None:
    # Should ignore return_properties
    retrieval_query = "WITH node MATCH (node)--(m) RETURN node, m LIMIT 10"
    match_query = get_match_query(
        return_properties=["name", "age"], retrieval_query=retrieval_query
    )
    expected = (
        "UNWIND $match_params AS match_param "
        "WITH match_param[0] AS match_id_value, match_param[1] AS score "
        "MATCH (node) "
        "WHERE node[$id_property] = match_id_value " + retrieval_query
    )
    assert match_query.strip() == expected.strip()


def test_match_query_with_custom_node_label() -> None:
    match_query = get_match_query(
        return_properties=["name", "age"], node_label="`MyNodeLabel`"
    )
    expected = (
        "UNWIND $match_params AS match_param "
        "WITH match_param[0] AS match_id_value, match_param[1] AS score "
        "MATCH (node:`MyNodeLabel`) "
        "WHERE node[$id_property] = match_id_value "
        "RETURN node {.name, .age} AS node, labels(node) AS nodeLabels, elementId(node) AS elementId, elementId(node) AS id, score "
    )
    assert match_query.strip() == expected.strip()


def test_weaviate_retriever_with_result_format_function(
    driver: MagicMock, neo4j_record: MagicMock, result_formatter: MagicMock
) -> None:
    query_text = "may thy knife chip and shatter"
    top_k = 5
    node_id_value = "node-test-id"
    node_match_score = 0.9

    wc = WClient(node_id_value=node_id_value, node_match_score=node_match_score)

    retriever = WeaviateNeo4jRetriever(
        driver=driver,
        client=wc,
        collection="dummy-collection",
        id_property_neo4j="sync_id",
        id_property_external="neo4j_id",
        result_formatter=result_formatter,
    )
    driver.execute_query.return_value = [
        [neo4j_record],
        None,
        None,
    ]

    records = retriever.search(query_text=query_text, top_k=top_k)

    assert records == RetrieverResult(
        items=[
            RetrieverResultItem(
                content="dummy-node", metadata={"score": 1.0, "node_id": 123}
            ),
        ],
        metadata={"__retriever": "WeaviateNeo4jRetriever"},
    )



"""The FuzzyMatchResolver merges nodes with same label
and similar textual properties (by default using the "name" property) based on RapidFuzz
for string matching.

If the resolution is intended to be applied only on some nodes, for instance nodes that
belong to a specific document, a "WHERE" query can be added. The only variable in the
query scope is "entity".

WARNING: this process is destructive, initial nodes are deleted and replaced
by the resolved ones, but all relationships are kept.
See apoc.refactor.mergeNodes documentation for more details.
"""

from neo4j_graphrag.experimental.components.resolver import (
    FuzzyMatchResolver,
)
from neo4j_graphrag.experimental.components.types import ResolutionStats

import neo4j


async def main(driver: neo4j.Driver) -> None:
    resolver = FuzzyMatchResolver(
        driver,
        # let's filter all entities that belong to a certain docId
        filter_query="WHERE (entity)-[:FROM_CHUNK]->(:Chunk)-[:FROM_DOCUMENT]->(doc:"
        "Document {id = 'docId'}",
        # optionally, change the properties used for resolution (default is "name")
        # resolve_properties=["name", "ssn"],
        # the similarity threshold (default is 0.8)
        # similarity_threshold=0.9
        # and the neo4j database where data is updated
        # neo4j_database="neo4j",
    )
    res: ResolutionStats = await resolver.run()
    print(res)



import neo4j
from neo4j_graphrag.indexes import create_fulltext_index

NEO4J_URL = "neo4j://localhost:7687"
NEO4J_AUTH = ("neo4j", "password")
FULLTEXT_INDEX_NAME = "fulltext_index"

driver = neo4j.GraphDatabase.driver(NEO4J_URL, auth=NEO4J_AUTH)

create_fulltext_index(
    driver, FULLTEXT_INDEX_NAME, label="Document", node_properties=["textProperty"]
)



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from unittest.mock import MagicMock, patch

import neo4j
import pytest
from neo4j.exceptions import CypherSyntaxError, Neo4jError
from neo4j_graphrag.exceptions import (
    RetrieverInitializationError,
    SchemaFetchError,
    SearchValidationError,
    Text2CypherRetrievalError,
)
from neo4j_graphrag.generation.prompts import Text2CypherTemplate
from neo4j_graphrag.llm import LLMResponse
from neo4j_graphrag.retrievers.text2cypher import Text2CypherRetriever, extract_cypher
from neo4j_graphrag.types import RetrieverResult, RetrieverResultItem


def test_t2c_retriever_initialization(driver: MagicMock, llm: MagicMock) -> None:
    with patch("neo4j_graphrag.retrievers.base.get_version") as mock_get_version:
        mock_get_version.return_value = ((5, 23, 0), False, False)
        Text2CypherRetriever(driver, llm, neo4j_schema="dummy-text")
        mock_get_version.assert_called_once()


@patch("neo4j_graphrag.retrievers.base.get_version")
@patch("neo4j_graphrag.retrievers.text2cypher.get_schema")
def test_t2c_retriever_schema_retrieval(
    get_schema_mock: MagicMock,
    mock_get_version: MagicMock,
    driver: MagicMock,
    llm: MagicMock,
) -> None:
    mock_get_version.return_value = ((5, 23, 0), False, False)
    Text2CypherRetriever(driver, llm)
    get_schema_mock.assert_called_once()


@patch("neo4j_graphrag.retrievers.base.get_version")
@patch("neo4j_graphrag.retrievers.text2cypher.get_schema")
def test_t2c_retriever_schema_retrieval_failure(
    get_schema_mock: MagicMock,
    mock_get_version: MagicMock,
    driver: MagicMock,
    llm: MagicMock,
) -> None:
    mock_get_version.return_value = ((5, 23, 0), False, False)
    get_schema_mock.side_effect = Neo4jError
    with pytest.raises(SchemaFetchError):
        Text2CypherRetriever(driver, llm)


@patch("neo4j_graphrag.retrievers.base.get_version")
def test_t2c_retriever_invalid_neo4j_schema(
    mock_get_version: MagicMock, driver: MagicMock, llm: MagicMock
) -> None:
    mock_get_version.return_value = ((5, 23, 0), False, False)
    with pytest.raises(RetrieverInitializationError) as exc_info:
        Text2CypherRetriever(
            driver=driver,
            llm=llm,
            neo4j_schema=42,  # type: ignore[arg-type, unused-ignore]
        )

    assert "neo4j_schema" in str(exc_info.value)
    assert "Input should be a valid string" in str(exc_info.value)


@patch("neo4j_graphrag.retrievers.base.get_version")
def test_t2c_retriever_invalid_search_query(
    mock_get_version: MagicMock, driver: MagicMock, llm: MagicMock
) -> None:
    mock_get_version.return_value = ((5, 23, 0), False, False)
    with pytest.raises(SearchValidationError) as exc_info:
        retriever = Text2CypherRetriever(
            driver=driver, llm=llm, neo4j_schema="dummy-text"
        )
        retriever.search(query_text=42)

    assert "query_text" in str(exc_info.value)
    assert "Input should be a valid string" in str(exc_info.value)


@patch("neo4j_graphrag.retrievers.base.get_version")
def test_t2c_retriever_invalid_search_examples(
    mock_get_version: MagicMock, driver: MagicMock, llm: MagicMock
) -> None:
    mock_get_version.return_value = ((5, 23, 0), False, False)
    with pytest.raises(RetrieverInitializationError) as exc_info:
        Text2CypherRetriever(
            driver=driver,
            llm=llm,
            neo4j_schema="dummy-text",
            examples=42,  # type: ignore[arg-type, unused-ignore]
        )

    assert "examples" in str(exc_info.value)
    assert "Initialization failed" in str(exc_info.value)


@patch("neo4j_graphrag.retrievers.base.get_version")
def test_t2c_retriever_happy_path(
    mock_get_version: MagicMock,
    driver: MagicMock,
    llm: MagicMock,
    neo4j_record: MagicMock,
) -> None:
    mock_get_version.return_value = ((5, 23, 0), False, False)
    t2c_query = "MATCH (n) RETURN n;"
    query_text = "may thy knife chip and shatter"
    neo4j_schema = "dummy-schema"
    examples = ["example-1", "example-2"]
    neo4j_database = "mydb"
    retriever = Text2CypherRetriever(
        driver=driver,
        llm=llm,
        neo4j_schema=neo4j_schema,
        examples=examples,
        neo4j_database=neo4j_database,
    )
    llm.invoke.return_value = LLMResponse(content=t2c_query)
    driver.execute_query.return_value = (
        [neo4j_record],
        None,
        None,
    )
    template = Text2CypherTemplate()
    prompt = template.format(
        schema=neo4j_schema,
        examples="\n".join(examples),
        query_text=query_text,
    )
    retriever.search(query_text=query_text)
    llm.invoke.assert_called_once_with(prompt)
    driver.execute_query.assert_called_once_with(
        query_=t2c_query,
        database_=neo4j_database,
        routing_=neo4j.RoutingControl.READ,
    )


@patch("neo4j_graphrag.retrievers.base.get_version")
def test_t2c_retriever_cypher_error(
    mock_get_version: MagicMock, driver: MagicMock, llm: MagicMock
) -> None:
    mock_get_version.return_value = ((5, 23, 0), False, False)
    t2c_query = "this is not a cypher query"
    neo4j_schema = "dummy-schema"
    examples = ["example-1", "example-2"]
    retriever = Text2CypherRetriever(
        driver=driver, llm=llm, neo4j_schema=neo4j_schema, examples=examples
    )
    retriever.llm.invoke.return_value = LLMResponse(content=t2c_query)
    query_text = "may thy knife chip and shatter"
    driver.execute_query.side_effect = CypherSyntaxError
    with pytest.raises(Text2CypherRetrievalError) as e:
        retriever.search(query_text=query_text)
    assert "Failed to get search result" in str(e)


@patch("neo4j_graphrag.retrievers.base.get_version")
def test_t2c_retriever_with_result_format_function(
    mock_get_version: MagicMock,
    driver: MagicMock,
    llm: MagicMock,
    neo4j_record: MagicMock,
    result_formatter: MagicMock,
) -> None:
    mock_get_version.return_value = ((5, 23, 0), False, False)
    retriever = Text2CypherRetriever(
        driver=driver, llm=llm, result_formatter=result_formatter
    )
    t2c_query = "MATCH (n) RETURN n;"
    retriever.llm.invoke.return_value = LLMResponse(content=t2c_query)
    query_text = "may thy knife chip and shatter"
    driver.execute_query.return_value = [
        [neo4j_record],
        None,
        None,
    ]

    records = retriever.search(query_text=query_text)

    assert records == RetrieverResult(
        items=[
            RetrieverResultItem(
                content="dummy-node", metadata={"score": 1.0, "node_id": 123}
            ),
        ],
        metadata={"cypher": t2c_query, "__retriever": "Text2CypherRetriever"},
    )


@patch("neo4j_graphrag.retrievers.text2cypher.extract_cypher")
@patch("neo4j_graphrag.retrievers.base.get_version")
def test_t2c_retriever_initialization_with_custom_prompt(
    mock_get_version: MagicMock,
    mock_extract_cypher: MagicMock,
    driver: MagicMock,
    llm: MagicMock,
    neo4j_record: MagicMock,
) -> None:
    mock_get_version.return_value = ((5, 23, 0), False, False)
    prompt = "This is a custom prompt. {query_text}"
    retriever = Text2CypherRetriever(driver=driver, llm=llm, custom_prompt=prompt)
    driver.execute_query.return_value = (
        [neo4j_record],
        None,
        None,
    )
    retriever.search(query_text="test")

    llm.invoke.assert_called_once_with("This is a custom prompt. test")


@patch("neo4j_graphrag.retrievers.text2cypher.extract_cypher")
@patch("neo4j_graphrag.retrievers.base.get_version")
def test_t2c_retriever_initialization_with_custom_prompt_and_schema_and_examples(
    mock_get_version: MagicMock,
    mock_extract_cypher: MagicMock,
    driver: MagicMock,
    llm: MagicMock,
    neo4j_record: MagicMock,
) -> None:
    mock_get_version.return_value = ((5, 23, 0), False, False)
    prompt = "This is a custom prompt. {query_text}"
    neo4j_schema = "dummy-schema"
    examples = ["example-1", "example-2"]

    retriever = Text2CypherRetriever(
        driver=driver,
        llm=llm,
        custom_prompt=prompt,
        neo4j_schema=neo4j_schema,
        examples=examples,
    )

    driver.execute_query.return_value = (
        [neo4j_record],
        None,
        None,
    )
    retriever.search(query_text="test")

    llm.invoke.assert_called_once_with("This is a custom prompt. test")


@patch("neo4j_graphrag.retrievers.text2cypher.extract_cypher")
@patch("neo4j_graphrag.retrievers.base.get_version")
def test_t2c_retriever_initialization_with_custom_prompt_and_schema_and_examples_for_prompt_params(
    mock_get_version: MagicMock,
    mock_extract_cypher: MagicMock,
    driver: MagicMock,
    llm: MagicMock,
    neo4j_record: MagicMock,
) -> None:
    mock_get_version.return_value = ((5, 23, 0), False, False)
    prompt = "This is a custom prompt. {query_text} {schema} {examples}"
    neo4j_schema = "dummy-schema"
    examples = ["example-1", "example-2"]

    retriever = Text2CypherRetriever(
        driver=driver,
        llm=llm,
        custom_prompt=prompt,
        neo4j_schema=neo4j_schema,
        examples=examples,
    )

    driver.execute_query.return_value = (
        [neo4j_record],
        None,
        None,
    )
    retriever.search(query_text="test")

    llm.invoke.assert_called_once_with(
        "This is a custom prompt. test dummy-schema example-1\nexample-2"
    )


@patch("neo4j_graphrag.retrievers.text2cypher.extract_cypher")
@patch("neo4j_graphrag.retrievers.base.get_version")
def test_t2c_retriever_initialization_with_custom_prompt_and_unused_schema_and_examples(
    mock_get_version: MagicMock,
    mock_extract_cypher: MagicMock,
    driver: MagicMock,
    llm: MagicMock,
    neo4j_record: MagicMock,
) -> None:
    mock_get_version.return_value = ((5, 23, 0), False, False)
    prompt = "This is a custom prompt. {query_text} {schema} {examples}"
    neo4j_schema = "dummy-schema"
    examples = ["example-1", "example-2"]

    retriever = Text2CypherRetriever(
        driver=driver,
        llm=llm,
        custom_prompt=prompt,
        neo4j_schema=neo4j_schema,
        examples=examples,
    )

    driver.execute_query.return_value = (
        [neo4j_record],
        None,
        None,
    )
    retriever.search(
        query_text="test",
        prompt_params={"schema": "another-dummy-schema", "examples": "another-example"},
    )

    llm.invoke.assert_called_once_with(
        "This is a custom prompt. test another-dummy-schema another-example"
    )


@patch("neo4j_graphrag.retrievers.text2cypher.extract_cypher")
@patch("neo4j_graphrag.retrievers.base.get_version")
def test_t2c_retriever_invalid_custom_prompt_type(
    mock_get_version: MagicMock,
    mock_extract_cypher: MagicMock,
    driver: MagicMock,
    llm: MagicMock,
) -> None:
    mock_get_version.return_value = ((5, 23, 0), False, False)
    with pytest.raises(RetrieverInitializationError) as exc_info:
        Text2CypherRetriever(
            driver=driver,
            llm=llm,
            custom_prompt=42,  # type: ignore[arg-type, unused-ignore]
        )

    assert "Input should be a valid string" in str(exc_info.value)


@patch("neo4j_graphrag.retrievers.text2cypher.extract_cypher")
@patch("neo4j_graphrag.retrievers.base.get_version")
def test_t2c_retriever_with_custom_prompt_prompt_params(
    mock_get_version: MagicMock,
    mock_extract_cypher: MagicMock,
    driver: MagicMock,
    llm: MagicMock,
    neo4j_record: MagicMock,
) -> None:
    mock_get_version.return_value = ((5, 23, 0), False, False)
    prompt = "This is a custom prompt. {query_text} {examples_custom}"
    query = "test"
    examples = ["example A", "example B"]

    retriever = Text2CypherRetriever(driver=driver, llm=llm, custom_prompt=prompt)
    driver.execute_query.return_value = (
        [neo4j_record],
        None,
        None,
    )
    retriever.search(query_text=query, prompt_params={"examples_custom": examples})

    llm.invoke.assert_called_once_with(
        """This is a custom prompt. test ['example A', 'example B']"""
    )


@patch("neo4j_graphrag.retrievers.text2cypher.extract_cypher")
@patch("neo4j_graphrag.retrievers.base.get_version")
def test_t2c_retriever_with_custom_prompt_bad_prompt_params(
    mock_get_version: MagicMock,
    mock_extract_cypher: MagicMock,
    driver: MagicMock,
    llm: MagicMock,
    neo4j_record: MagicMock,
) -> None:
    mock_get_version.return_value = ((5, 23, 0), False, False)
    prompt = "This is a custom prompt. {query_text} {examples}"
    query = "test"
    examples = ["example A", "example B"]

    retriever = Text2CypherRetriever(driver=driver, llm=llm, custom_prompt=prompt)
    driver.execute_query.return_value = (
        [neo4j_record],
        None,
        None,
    )
    retriever.search(
        query_text=query,
        prompt_params={
            "examples": examples,
            "bad_param": "this should not be present in template.",
        },
    )

    llm.invoke.assert_called_once_with(
        """This is a custom prompt. test ['example A', 'example B']"""
    )


@patch("neo4j_graphrag.retrievers.text2cypher.extract_cypher")
@patch("neo4j_graphrag.retrievers.base.get_version")
@patch("neo4j_graphrag.retrievers.text2cypher.get_schema")
def test_t2c_retriever_with_custom_prompt_and_schema(
    get_schema_mock: MagicMock,
    mock_get_version: MagicMock,
    mock_extract_cypher: MagicMock,
    driver: MagicMock,
    llm: MagicMock,
    neo4j_record: MagicMock,
) -> None:
    mock_get_version.return_value = ((5, 23, 0), False, False)
    prompt = "This is a custom prompt. {query_text} {schema}"
    query = "test"

    driver.execute_query.return_value = (
        [neo4j_record],
        None,
        None,
    )

    retriever = Text2CypherRetriever(driver=driver, llm=llm, custom_prompt=prompt)
    retriever.search(
        query_text=query,
        prompt_params={},
    )

    get_schema_mock.assert_not_called()
    llm.invoke.assert_called_once_with("""This is a custom prompt. test """)


@pytest.mark.parametrize(
    "description, cypher_query, expected_output",
    [
        ("No changes", "MATCH (n) RETURN n;", "MATCH (n) RETURN n;"),
        (
            "Surrounded by backticks",
            "Cypher query: ```MATCH (n) RETURN n;```",
            "MATCH (n) RETURN n;",
        ),
        (
            "Spaces in label",
            "Cypher query: ```MATCH (n: Label With Spaces ) RETURN n;```",
            "MATCH (n:`Label With Spaces`) RETURN n;",
        ),
        (
            "No spaces in label",
            "Cypher query: ```MATCH (n: LabelWithNoSpaces ) RETURN n;```",
            "MATCH (n: LabelWithNoSpaces ) RETURN n;",
        ),
        (
            "Backticks in label",
            "Cypher query: ```MATCH (n: `LabelWithBackticks` ) RETURN n;```",
            "MATCH (n: `LabelWithBackticks` ) RETURN n;",
        ),
        (
            "Spaces in property key",
            "Cypher query: ```MATCH (n: { prop 1: 1, prop 2: 2 }) RETURN n;```",
            "MATCH (n: { `prop 1`: 1, `prop 2`: 2 }) RETURN n;",
        ),
        (
            "No spaces in property key",
            "Cypher query: ```MATCH (n: { prop1: 1, prop2: 2 }) RETURN n;```",
            "MATCH (n: { prop1: 1, prop2: 2 }) RETURN n;",
        ),
        (
            "Backticks in property key",
            "Cypher query: ```MATCH (n: { `prop 1`: 1, `prop 2`: 2 }) RETURN n;```",
            "MATCH (n: { `prop 1`: 1, `prop 2`: 2 }) RETURN n;",
        ),
        (
            "Spaces in relationship type",
            "Cypher query: ```MATCH (n)-[: Relationship With Spaces ]->(m) RETURN n, m;```",
            "MATCH (n)-[:`Relationship With Spaces`]->(m) RETURN n, m;",
        ),
        (
            "No spaces in relationship type",
            "Cypher query: ```MATCH (n)-[ : RelationshipWithNoSpaces ]->(m) RETURN n, m;```",
            "MATCH (n)-[ : RelationshipWithNoSpaces ]->(m) RETURN n, m;",
        ),
        (
            "Backticks in relationship type",
            "Cypher query: ```MATCH (n)-[ : `RelationshipWithBackticks` ]->(m) RETURN n, m;```",
            "MATCH (n)-[ : `RelationshipWithBackticks` ]->(m) RETURN n, m;",
        ),
    ],
)
def test_extract_cypher(
    description: str, cypher_query: str, expected_output: str
) -> None:
    assert (
        extract_cypher(cypher_query) == expected_output
    ), f"Failed test case: {description}"
